import hmac
import base64
import hashlib
import json
import datetime
import requests
from urllib import parse
import urllib.parse
from datetime import datetime as dt
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general
# Requests will use simplejson if available.
try:
    from simplejson.errors import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

class RestApiRequest(object):

    def __init__(self):
        self.method = ""
        self.url = ""
        self.host = ""
        self.post_body = ""
        self.header = dict()
        self.json_parser = None

class UrlParamsBuilder(object):

    def __init__(self):
        self.param_map = dict()
        self.post_map = dict()
        self.post_list = list()

    def put_url(self, name, value):
        if value is not None:
            if isinstance(value, (list, dict)):
                self.param_map[name] = value
            else:
                self.param_map[name] = str(value)

    def put_post(self, name, value):
        if value is not None:
            if isinstance(value, (list, dict)):
                self.post_map[name] = value
            else:
                self.post_map[name] = str(value)

    def build_url(self):
        if len(self.param_map) == 0:
            return ""
        encoded_param = urllib.parse.urlencode(self.param_map)
        return "?" + encoded_param

    def build_url_to_json(self):
        return json.dumps(self.param_map)

class huobi(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.huobi.pro"
        self.domain_url_2 = "https://api-aws.huobi.pro"
        self.prefix = "/api/v4"
        self.category = kwargs["category"] if "category" in kwargs else "spot"
        self.recv_window = kwargs["recv_window"] if "recv_window" in kwargs else "5000"
        self.account_type = kwargs["account_type"] if "account_type" in kwargs else "UNIFIED"
        self.max_retries = 5
        self.session = requests.Session()

    def _getSymbol(self, symbol):
        return "".join([x.lower() for x in symbol.split("/")])
    
    def _getUserSymbol(self, symbol):
        raise NotImplementedError("method not implemented") 
    
    def utc_now(self):
        return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')

    ## Auth 
    def create_signature(self, method, url, builder):
        if self.key is None or self.secret is None or self.key == "" or self.secret == "":
            raise Exception("API key and secret key are required")

        timestamp = self.utc_now()
        builder.put_url("AccessKeyId", self.key)
        builder.put_url("SignatureVersion", "2")
        builder.put_url("SignatureMethod", "HmacSHA256")
        builder.put_url("Timestamp", timestamp)

        host = urllib.parse.urlparse(url).hostname
        path = urllib.parse.urlparse(url).path

        keys = sorted(builder.param_map.keys())
        qs0 = '&'.join(['%s=%s' % (key, parse.quote(builder.param_map[key], safe='')) for key in keys])
        payload0 = '%s\n%s\n%s\n%s' % (method, host, path, qs0)
        dig = hmac.new(self.secret.encode('utf-8'), msg=payload0.encode('utf-8'), digestmod=hashlib.sha256).digest()
        s = base64.b64encode(dig).decode()
        builder.put_url("Signature", s)

    def __create_request_by_get_with_signature(self, url, builder):
        request = RestApiRequest()
        request.method = "GET"
        request.host = self.domain_url
        self.create_signature(request.method, request.host + url, builder)
        request.header.update({"Content-Type": "application/x-www-form-urlencoded"})
        request.url = url + builder.build_url()
        return request
    
    def __create_request_by_post_with_signature(self, url, builder):
        request = RestApiRequest()
        request.method = "POST"
        request.host = self.domain_url
        self.create_signature(request.method, request.host + url, builder)
        request.header.update({'Content-Type': 'application/json'})
        if (len(builder.post_list)):  # specify for case : /v1/order/batch-orders
            request.post_body = builder.post_list
        else:
            request.post_body = builder.post_map
        request.url = url + builder.build_url()
        return request
    
    def create_request(self, method, url, params):
        builder = UrlParamsBuilder()
        if params and len(params):
            if method == "GET" or method == "GET-UNSIGNED":
                for key, value in params.items():
                    builder.put_url(key, value)
            elif method == "POST" or method == "POST-UNSIGNED":
                for key, value in params.items():
                    builder.put_post(key, value)
            else:
                raise Exception("[error] undefined HTTP method")

        if method =="GET-UNSIGNED":
            request = self.__create_request_by_get(url, builder)
        elif method == "GET":
            request = self.__create_request_by_get_with_signature(url, builder)
        elif method == "POST":
            request = self.__create_request_by_post_with_signature(url, builder)
        else:
            raise Exception(method + "  is invalid http method")

        return request
    
    def _signedRequest(self, method, url, params, is_checked=False):
        request = self.create_request(method, url, params)
        if request.method == "GET":
            response = self.session.get(request.host + request.url, headers=request.header)
            if is_checked is True:
                return response.text
            dict_data = json.loads(response.text)
            return dict_data

        elif request.method == "POST":
            response = self.session.post(request.host + request.url, data=json.dumps(request.post_body), headers=request.header)
            dict_data = json.loads(response.text)
            return dict_data
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + self.prefix + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")

    ## parsers
    def _parseBalance(self, balData):
        parsedBal = {"free": {}, "total": {}}
        if balData["status"] != "ok": raise Exception(balData["err-msg"])
        for balDataEle in balData["data"]["list"]:
            if balDataEle["type"] == "trade":
                parsedBal["free"][balDataEle["currency"].upper()] = float(balDataEle["balance"])
            if balDataEle["type"] == "frozen":
                parsedBal["total"][balDataEle["currency"].upper()] = float(balDataEle["balance"]) + parsedBal["free"][balDataEle["currency"].upper()]
        return parsedBal
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("ts", None)
        parsedData["nonce"] = orderbookData.get("u", None)
        if "b" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["b"]]
        else: parsedData["bids"] = []
        if "a" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["a"]]
        else: parsedData["asks"] = []
        return parsedData
    
    def _parseCreateOrder(self, order):
        parsedOrder = {}
        if order["status"] != "ok": raise Exception(order["err-msg"])
        parsedOrder["id"] = order["data"]
        parsedOrder["symbol"] = None
        parsedOrder["price"] = None
        parsedOrder["amount"] = None
        parsedOrder["side"] = None
        parsedOrder["timestamp"] = None
        parsedOrder["status"] = None
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    '''
    {'retCode': 0, 'retMsg': 'OK', 'result': {'orderId': 'f3708d4e-24d6-4528-9037-764d03610479', 'orderLinkId': ''}, 
    'retExtInfo': {}, 'time': 1702055115208}
    '''
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order["orderId"]
        parsedOrder["symbol"] = order.get('symbol', None)
        parsedOrder["price"] = order.get('price', None)
        parsedOrder["amount"] = order.get('amount', None)
        parsedOrder["side"] = "buy" if order["type"].startswith("buy") else "sell"
        parsedOrder["timestamp"] = int(order["created-at"])
        parsedOrder["status"] = order.get('state', None)
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder

    def _parseOpenOrders(self, orders):
        parsedOrderList = []
        if order["status"] != "ok": raise Exception(order["err-msg"])
        for order in orders['data']:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList

    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["id"] = trade.get("order-id", None)
            parsedTrade["tradeId"] = trade.get("trade-id", None)
            parsedTrade["symbol"] = trade.get("symbol", None)
            parsedTrade["side"] = "buy" if trade["type"].startswith("buy") else "sell"
            parsedTrade["price"] = trade.get("price", None)
            parsedTrade["amount"] = trade.get("filled-amount", None)
            parsedTrade["takerOrMaker"] = trade["role"]
            if "created-at" in trade:
                parsedTrade["timestamp"] = trade["created-at"]
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            parsedTrade["feeCurrency"] = trade["fee-currency"]
            parsedTrade["fee"] = trade["filled-fees"]
            return parsedTrade
        parsedTradeList = []
        if trades["status"] != "ok": raise Exception(trades["err-msg"])
        for trade in trades["data"]:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList


    ## Exchange functions 

    # https://www.gate.io/docs/developers/apiv4/en/#create-a-futures-order
    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            if type == "limit":
                return self.create_limit_order(symbol, amount, side, price, params={})
            elif type == "market":
                return self.create_market_order(symbol, amount, side, params={})
        except Exception as e:
            raise e
    
    def create_limit_order(self, symbol, amount, side, price, params={}):
        body = {
            "account-id": self.fetch_accounts(),
            "amount": amount,
            "price": price,
            "source": "spot-api",
            "symbol": self._getSymbol(symbol),
            "type": "buy-limit" if side == "buy" else "sell-limit"
        }
        apiUrl = "/v1/order/orders/place"
        try:
            resp = self._signedRequest(method='POST', url=apiUrl, params=body)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e
        
    def create_market_order(self, symbol, amount, side, params={}):
        body = {
            "account-id": self.fetch_accounts(),
            "amount": amount,
            "source": "spot-api",
            "symbol": self._getSymbol(symbol),
            "type": "buy-market" if side == "buy" else "sell-market"
        }
        apiUrl = "/v1/order/orders/place"
        try:
            resp = self._signedRequest(method='POST', url=apiUrl, params=body)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        raise NotImplementedError("method not implemented")

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/v1/order/openOrders"
        params = {"symbol": self._getSymbol(symbol)}
        try:
            resp = self._signedRequest(method='GET', url=apiUrl, params=params)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None, params={}):
        apiUrl = f"/v1/order/orders/{id}/submitcancel"
        body = {
            "order_id" : id
        }
        try:
            resp = self._signedRequest(method='POST', url=apiUrl, params=body)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        raise NotImplementedError("method not implemented")
    
    def fetch_accounts(self):
        apiUrl = '/v1/account/accounts'
        try:
            resp = self._signedRequest(method='GET', url=apiUrl, params={})
            if resp['status'] != "ok": raise Exception("error fetching accounts")
            return resp['data'][0]['id']
        except Exception as e:
            raise e
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        accountId = self.fetch_accounts()
        apiUrl = f"/v1/account/accounts/{accountId}/balance"
        try:
            resp = self._signedRequest(method='GET', url=apiUrl, params={})
            return self._parseBalance(resp)
        except Exception as e:
            raise e
    
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        apiUrl = '/v1/order/matchresults'
        queryParams = {"symbol": self._getSymbol(symbol)}
        if since:
            queryParams["start-time"] = since
        if limit:
            queryParams["size"] = limit
        if 'endTime' in params:
            queryParams["end-time"] = params['endTime']
        try:
            resp = self._signedRequest(method='GET', url=apiUrl, params=queryParams)
            return self._parseTrades(resp)
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        raise NotImplementedError("method not implemented")
        
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("Subclasses must implement this method")


import hmac
import base64
import hashlib
import json
import time
import requests
from datetime import datetime as dt
from ttxt.base import baseFuturesExchange
from ttxt.types import baseTypes
# Requests will use simplejson if available.
try:
    from simplejson.errors import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

'''
kwards = {
    "category": "",
    "recv_window": ""
}
'''
class bybitFutures(baseFuturesExchange.BaseFuturesExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url_testnet = "https://api-testnet.bybit.com"
        self.domain_url = "https://api.bybit.com"
        self.domain_url_2 = "https://api.bytick.com"
        self.prefix = "/api/v4"
        self.category = kwargs["category"] if "category" in kwargs else "linear"
        self.recv_window = kwargs["recv_window"] if "recv_window" in kwargs else "5000"
        self.account_type = kwargs["account_type"] if "account_type" in kwargs else "UNIFIED"
        self.max_retries = 5
        self.symbolLotSize = {}

    def _getSymbol(self, symbol):
        return symbol.replace("/", "")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("_", "/") # TODO: change this 
    
    def _checkAndSetMinQty(self, ticker):
        if ticker in self.symbolLotSize and self.symbolLotSize[ticker] != {}:
            return 
        info = self.getInstrumentInfo(symbol=ticker)
        self.symbolLotSize = info["result"]["list"][0]["lotSizeFilter"]

    ## Auth 
    def get_expire(self):
        return int((time.time() + 1) * 1000)  # websockets use seconds
    
    def generate_timestamp(self):
        """
        Return a millisecond integer timestamp.
        """
        return int(time.time() * 10**3)

    def generate_signature(self, param_str):
        hash = hmac.new(
            bytes(self.secret, "utf-8"),
            param_str.encode("utf-8"),
            hashlib.sha256,
        )
        return hash.hexdigest()

    def prepare_payload(self, method, parameters):
        """
        Prepares the request payload and validates parameter value types.
        """

        def cast_values():
            string_params = [
                "qty",
                "price",
                "triggerPrice",
                "takeProfit",
                "stopLoss",
            ]
            integer_params = ["positionIdx"]
            for key, value in parameters.items():
                if key in string_params:
                    if type(value) != str:
                        parameters[key] = str(value)
                elif key in integer_params:
                    if type(value) != int:
                        parameters[key] = int(value)

        if method == "GET":
            payload = "&".join(
                [
                    str(k) + "=" + str(v)
                    for k, v in sorted(parameters.items())
                    if v is not None
                ]
            )
            return payload
        else:
            cast_values()
            return json.dumps(parameters)

    def _auth(self, payload, recv_window, timestamp):
        """
        Prepares authentication signature per Bybit API specifications.
        """

        if self.key is None or self.secret is None:
            raise PermissionError("Authenticated endpoints require keys.")
        param_str = str(timestamp) + self.key + str(recv_window) + payload
        return self.generate_signature(param_str)

    @staticmethod
    def _verify_string(self,params, key):
        if key in params:
            if not isinstance(params[key], str):
                return False
            else:
                return True
        return True

    def _signedRequest(self, method=None, path=None, query=None, auth=False):
        """
        Submits the request to the API.

        Notes
        -------------------
        We use the params argument for the GET method, and data argument for
        the POST method. Dicts passed to the data argument must be
        JSONified prior to submitting request.

        """
        path = self.domain_url + path
        if query is None:
            query = {}
        # Store original recv_window.
        recv_window = self.recv_window
        # Bug fix: change floating whole numbers to integers to prevent
        # auth signature errors.
        if query is not None:
            for i in query.keys():
                if isinstance(query[i], float) and query[i] == int(query[i]):
                    query[i] = int(query[i])
        # Send request and return headers with body. Retry if failed.
        retries_attempted = self.max_retries
        req_params = None
        while True:
            retries_attempted -= 1
            if retries_attempted < 0:
                raise Exception(
                    "Bad Request. Retries exceeded maximum."
                )
            retries_remaining = f"{retries_attempted} retries remain."
            req_params = self.prepare_payload(method, query)
            # Authenticate if we are using a private endpoint.
            if auth:
                # Prepare signature.
                timestamp = self.generate_timestamp()
                signature = self._auth(
                    payload=req_params,
                    recv_window=recv_window,
                    timestamp=timestamp,
                )
                headers = {
                    "Content-Type": "application/json",
                    "X-BAPI-API-KEY": self.key,
                    "X-BAPI-SIGN": signature,
                    "X-BAPI-SIGN-TYPE": "2",
                    "X-BAPI-TIMESTAMP": str(timestamp),
                    "X-BAPI-RECV-WINDOW": str(recv_window),
                }
            else:
                headers = {}
            if method == "GET":
                try:
                    if req_params:
                        client = requests.Session()
                        r = client.prepare_request(requests.Request(method, path, headers=headers))
                        #esp = client.send(r, timeout=60) 
                        resp = requests.get(path + f"?{req_params}", headers=headers)
                    else:
                        r = requests.get(path, headers=headers)
                    return resp.json()
                except Exception as e:
                    raise e
            if method == "POST":
                r = requests.post( path, data=req_params, headers=headers)
                return r.json()
            if method == "DELETE":
                r = requests.delete( path, data=req_params, headers=headers)
                return r.json()
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + self.prefix + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")

    ## parsers
    '''
    {"retCode":0,"retMsg":"OK","result":
    {"list":[{"totalEquity":"233.33780771","accountIMRate":"0","totalMarginBalance":"233.07030533",
    "totalInitialMargin":"0","accountType":"UNIFIED","totalAvailableBalance":"233.07030533",
    "accountMMRate":"0","totalPerpUPL":"0","totalWalletBalance":"233.07030533",
    "accountLTV":"0","totalMaintenanceMargin":"0",
    "coin":[{"availableToBorrow":"","bonus":"0","accruedInterest":"0","availableToWithdraw":"0.00000978",
        "totalOrderIM":"0","equity":"0.00000978","totalPositionMM":"0","usdValue":"0.02315849",
        "unrealisedPnl":"0","collateralSwitch":true,"spotHedgingQty":"0","borrowAmount":"0.000000000000000000",
        "totalPositionIM":"0","walletBalance":"0.00000978","cumRealisedPnl":"-0.00000598","locked":"0",
        "marginCollateral":true,"coin":"ETH"},
        {"availableToBorrow":"","bonus":"0","accruedInterest":"",
        "availableToWithdraw":"0.004","totalOrderIM":"0","equity":"0.004","totalPositionMM":"0",
        "usdValue":"0.00481953","unrealisedPnl":"0","collateralSwitch":false,"spotHedgingQty":"0",
        "borrowAmount":"0.000000000000000000","totalPositionIM":"0","walletBalance":"0.004",
        "cumRealisedPnl":"0","locked":"0","marginCollateral":false,"coin":"XCAD"},
        {"availableToBorrow":"","bonus":"0","accruedInterest":"0","availableToWithdraw":"232.52736215","totalOrderIM":"0","equity":"232.52736215","totalPositionMM":"0","usdValue":"232.56261562","unrealisedPnl":"0","collateralSwitch":true,"spotHedgingQty":"0","borrowAmount":"0.000000000000000000","totalPositionIM":"0","walletBalance":"232.52736215","cumRealisedPnl":"-0.07024754","locked":"0","marginCollateral":true,"coin":"USDT"},{"availableToBorrow":"","bonus":"0","accruedInterest":"0","availableToWithdraw":"24.0062","totalOrderIM":"0","equity":"24.0062","totalPositionMM":"0","usdValue":"0.74721405","unrealisedPnl":"0","collateralSwitch":true,"spotHedgingQty":"0","borrowAmount":"0.000000000000000000","totalPositionIM":"0","walletBalance":"24.0062","cumRealisedPnl":"-1.7438","locked":"0","marginCollateral":true,"coin":"GALA"}]}]},"retExtInfo":{},"time":1702038150602}
    '''
    # balData = {"free": {"USDT": 1010, "BTC": ""}, "total": {"USDT": 1010, "BTC": 0}, "unrealisedPnl": {"USDT": 1010, "BTC": 0}}
    def _parseBalance(self, balData):
        parsedBal = {"free": {}, "total": {}, "unrealisedPnl": {}}
        balDataResult = balData.get("result", None)
        balDatList = balDataResult.get("list", [])
        if len(balDatList) == 0:
            return parsedBal
        balDatList = balDatList[0].get("coin", [])
        if len(balDatList) == 0:
            return parsedBal
        for balDataEle in balDatList:
            parsedBal["free"][balDataEle["coin"]] = balDataEle.get("walletBalance", None)
            parsedBal["total"][balDataEle["coin"]] = balDataEle.get("equity", None)
            parsedBal["unrealisedPnl"][balDataEle["coin"]] = balDataEle.get("unrealisedPnl", None)
        return parsedBal
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("ts", None)
        parsedData["nonce"] = orderbookData.get("u", None)
        if "b" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["b"]]
        else: parsedData["bids"] = []
        if "a" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["a"]]
        else: parsedData["asks"] = []
        return parsedData

    '''
    {'retCode': 0, 'retMsg': 'OK', 'result': {'orderId': 'f3708d4e-24d6-4528-9037-764d03610479', 'orderLinkId': ''}, 
    'retExtInfo': {}, 'time': 1702055115208}
    '''
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order["orderId"]
        parsedOrder["symbol"] = self._getUserSymbol(order["contract"])
        parsedOrder["price"] = float(order["price"])
        parsedOrder["amount"] = float(order["size"])
        parsedOrder["side"] = "buy" if parsedOrder["amount"] > 0 else "sell"
        parsedOrder["timestamp"] = float(order["create_time"])
        parsedOrder["status"] = order["status"]
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    def _parseCreateorder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order["orderId"]
        parsedOrder["symbol"] = None
        parsedOrder["price"] = None
        parsedOrder["amount"] = None
        parsedOrder["side"] = None
        parsedOrder["timestamp"] = None
        parsedOrder["status"] = None
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder

    def _parseOpenOrders(self, orders):
        parsedOrderList = []
        for order in orders:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList

    '''
    {'retCode': 0, 'retMsg': 'OK', 'result': {'nextPageCursor': 'f3708d4e-24d6-4528-9037-764d03610479%3A1702055115208%2Cf3708d4e-24d6-4528-9037-764d03610479%3A1702055115208', 'category': 'linear', 'list': [{'symbol': 'BTCUSDT', 'orderType': 'Limit', 'orderLinkId': '', 'slLimitPrice': '0', 'orderId': 'f3708d4e-24d6-4528-9037-764d03610479', 'cancelType': 'UNKNOWN', 'avgPrice': '43832.1', 'stopOrderType': '', 'lastPriceOnCreated': '43832', 'orderStatus': 'Filled', 'takeProfit': '', 'cumExecValue': '43.8321', 'tpslMode': 'UNKNOWN', 'smpType': 'None', 'triggerDirection': 0, 'blockTradeId': '', 'isLeverage': '', 'rejectReason': 'EC_NoError', 'price': '43920', 'orderIv': '', 'createdTime': '1702055115208', 'tpTriggerBy': '', 'positionIdx': 0, 'timeInForce': 'GTC', 'leavesValue': '0', 'updatedTime': '1702055115209', 'side': 'Buy', 'smpGroup': 0, 'triggerPrice': '', 'tpLimitPrice': '0', 'cumExecFee': '0.02410766', 'leavesQty': '0', 'slTriggerBy': '', 'closeOnTrigger': False, 'placeType': '', 'cumExecQty': '0.001', 'reduceOnly': False, 'qty': '0.001', 'stopLoss': '', 'smpOrderId': '', 'triggerBy': ''}]}, 'retExtInfo': {}, 'time': 1702055320777}
    '''
    def _parseFetchedOrder(self, order):
        pass

    ## Exchange functions 

    # https://www.gate.io/docs/developers/apiv4/en/#create-a-futures-order
    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            ticker = self._getSymbol(symbol)
            self._checkAndSetMinQty(ticker=ticker)
            # TODO: check multiple for qty
            body = {
                "category": self.category,
                "symbol": ticker,
                "isLeverage": params["isLeverage"] if "isLeverage" in params else 1,  # 1 is for margin, 0 for spot
                "side": side.capitalize(),
                "orderType": type,
                "qty": float(amount),
                "timeInForce": params["timeInForce"] if "timeInForce" in params else "GTC",
                "positionIdx": params["positionIdx"] if "positionIdx" in params else 0,
            }
            body.update(params) 
            if type == "limit":
               body["price"] = float(price)
            apiUrl = "/v5/order/create"
            response = self._signedRequest(method='POST', path=apiUrl, query=body, auth=True)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        apiUrl = "/v5/order/realtime"
        queryParams = {"category": self.category, "orderId": id}
        try:
            resp = resp = self._signedRequest(method='GET', path=apiUrl, query=queryParams, auth=True)
            return resp #self._parseOrder(resp)
        except Exception as e:
            raise e

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/v5/order/realtime"
        queryParams = {"category": self.category}
        try:
            resp = resp = self._signedRequest(method='GET', path=apiUrl, query=queryParams, auth=True)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None, params={}):
        apiUrl = "/v5/order/cancel"
        queryParams = {"category": self.category, "orderId": id, "symbol": self._getSymbol(symbol)}
        try:
            resp = self._signedRequest(method='POST', path=apiUrl, query=queryParams, auth=True)
            return resp #self._parseOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        apiUrl = "/v5/market/tickers"
        queryParams = {"symbol": self._getSymbol(symbol), "category": self.category}
        try:
            resp = self._signedRequest(method='GET', path=apiUrl, query=queryParams)
            return resp # parse this response into Ticker
        except Exception as e:
            raise e
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/v5/account/wallet-balance'
        queryParams = {"accountType": self.account_type}
        try:
            resp = self._signedRequest(method='GET', path=apiUrl, query=queryParams, auth=True)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
    
    def set_leverage(self, leverage, symbol, params={}):
        apiUrl = "/v5/position/set-leverage"
        queryParams = {"category": self.category, "symbol": self._getSymbol(symbol), "buyLeverage": str(leverage), "sellLeverage": str(leverage)}
        try:
            resp = self._signedRequest(method='POST', path=apiUrl, query=queryParams, auth=True)
            return resp
        except Exception as e:
            raise e

    ## extra functions
    def getInstrumentInfo(self, symbol):
        apiUrl = "/v5/market/instruments-info"
        queryParams = {"symbol": self._getSymbol(symbol), "category": self.category}
        try:
            resp = self._signedRequest(method='GET', path=apiUrl, query=queryParams)
            return resp # parse this response into Ticker
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        apiUrl = "/v5/market/orderbook"
        queryParams = {"symbol": self._getSymbol(symbol), "category": self.category}
        if limit is not None:
            queryParams["limit"] = int(limit)
        queryParams.update(params)
        try:
            resp = self._signedRequest(method='GET', path=apiUrl, query=queryParams)
            return self._parseOrderbook(resp)
        except Exception as e:
            raise e
        
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("Subclasses must implement this method")


from enum import Enum, unique
import hmac
import base64
import hashlib
import json
import time
import requests
import datetime
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general

@unique
class Auth(Enum):
    NONE = 1
    KEYED = 2
    SIGNED = 3

# memo = UID => password
class bitmart(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, password, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.password = password
        self.domain_url = 'https://api-cloud.bitmart.com'  # host + prefix
        # http header
        self.CONTENT_TYPE = 'Content-Type'
        self.USER_AGENT = 'User-Agent'
        self.X_BM_KEY = 'X-BM-KEY'
        self.X_BM_SIGN = 'X-BM-SIGN'
        self.X_BM_TIMESTAMP = 'X-BM-TIMESTAMP'
        self.__version__ = "1.0.1"
        # http header
        self.APPLICATION_JSON = 'application/json'
        self.VERSION = 'bitmart-python-sdk-api/'
        self.GET = "GET"
        self.POST = "POST"
        self.DELETE = "DELETE"
        self.TIMEOUT = (5,10)

    def _getSymbol(self, symbol):
        return symbol.replace("/", "_")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("_", "/")

    ## Auth 
    def sign(self, message):
        mac = hmac.new(bytes(self.secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        return mac.hexdigest()

    # timestamp + "#" + memo + "#" + queryString
    def pre_substring(self, timestamp, body):
        return f'{str(timestamp)}#{self.password}#{body}'

    def get_header(self, sign, timestamp):
        header = dict()
        header[self.CONTENT_TYPE] = self.APPLICATION_JSON
        header[self.USER_AGENT] = self.VERSION + self.__version__

        if self.key:
            header[self.X_BM_KEY] = self.key
        if sign:
            header[self.X_BM_SIGN] = sign
        if timestamp:
            header[self.X_BM_TIMESTAMP] = str(timestamp)

        return header

    def parse_params_to_str(self, params):
        url = '?'
        for key, value in params.items():
            url = url + str(key) + '=' + str(value) + '&'

        return url[0:-1]

    def get_timestamp(self):
        return str(datetime.datetime.now().timestamp() * 1000).split('.')[0]

    ## Requests 
    def _signedRequest(self, method, request_path, params, auth):
        if method == self.GET or method == self.DELETE:
            url = self.domain_url + request_path + self.parse_params_to_str(params)
        else:
            url = self.domain_url + request_path

        # set body
        body = json.dumps(params) if method == self.POST else ""

        # set header
        if auth == Auth.NONE:
            header = self.get_header(sign=None, timestamp=None)
        elif auth == Auth.KEYED:
            header = self.get_header(sign=None, timestamp=None)
        else:
            timestamp = self.get_timestamp()
            sign = self.sign(self.pre_substring(timestamp, str(body)))
            header = self.get_header(sign, timestamp)

        # send request
        response = None
        if method == self.GET:
            response = requests.get(url, headers=header, timeout=self.TIMEOUT)
        elif method == self.POST:
            response = requests.post(url, data=body, headers=header, timeout=self.TIMEOUT)
        elif method == self.DELETE:
            response = requests.delete(url, headers=header, timeout=self.TIMEOUT)

        # exception handle
        if not str(response.status_code) == '200':
            raise Exception(response)
        try:
            res_header = response.headers
            r = dict()
            try:
                r['Remaining'] = res_header['X-BM-RateLimit-Remaining']
                r['Limit'] = res_header['X-BM-RateLimit-Limit']
                r['Reset'] = res_header['X-BM-RateLimit-Reset']
            except:
                pass
            return response.json()

        except ValueError:
            raise Exception('Invalid Response: %s' % response.text)

    ## parsers
    def _parseBalance(self, balData):
        parsedBal = {"free": {}, "total": {}}
        if balData['code'] != 1000: raise Exception(balData['message'])
        for bal in balData['data']['wallet']:
            parsedBal["free"][bal["currency"]] = float(bal["available"])
            parsedBal["total"][bal["currency"]] = float(bal['frozen']) + parsedBal["free"][bal["currency"]]
        return parsedBal
    
    def _parseCreateOrder(self, orderData):
        parsedOrder = {}
        if orderData['code'] != 1000: raise Exception(orderData['message'])
        parsedOrder["orderId"] = orderData["data"]["order_id"]
        parsedOrder["symbol"] = None
        parsedOrder["price"] = None
        parsedOrder["amount"] = None
        parsedOrder["side"] = None
        parsedOrder["timestamp"] = None
        parsedOrder["status"] = None
        parsedOrder["orderJson"] = json.dumps(orderData)
        return parsedOrder
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("current", None)
        parsedData["nonce"] = orderbookData.get("id", None)
        if "bids" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["bids"]]
        else: parsedData["bids"] = []
        if "asks" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["asks"]]
        else: parsedData["asks"] = []
        return parsedData
    
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('orderId', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["symbol"]) if "symbol" in order else None
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["size"]) if "size" in order else None
        parsedOrder["side"] = order.get('side', None)
        parsedOrder["timestamp"] = int(float(order["createTime"])) if "createTime" in order else None
        parsedOrder["status"] = order.get('state', None)
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseOpenOrders(self, orders):
        if orders['code'] != 1000: raise Exception(orders['message'])
        parsedOrderList = []
        for order in orders["data"]:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList

    # cancel request doesnt return an order id in response
    def _parseCancelOrder(self, order):
        if order['code'] != 1000: raise Exception(order['message'])
        return order['data']
    
    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["id"] = trade.get("orderId", None)
            parsedTrade["tradeId"] = trade.get("tradeId", None)
            parsedTrade["side"] = trade.get("side", None)
            parsedTrade["price"] = trade.get("price", None)
            parsedTrade["amount"] = trade.get("size", None)
            parsedTrade["symbol"] = self._getUserSymbol(trade["symbol"])
            parsedTrade["takerOrMaker"] = trade.get("tradeRole", None)
            if "updateTime" in trade:
                parsedTrade["timestamp"] = int(trade["updateTime"])
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            return parsedTrade
        parsedTradeList = []
        if trades['code'] != 1000: raise Exception(trades['message'])
        for trade in trades['data']:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList

    ## Exchange functions 

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            if type == "market":
                body = {
                    'symbol': self._getSymbol(symbol),
                    'side': side,
                    'type': 'market',
                    'size': amount,
                    'notional': amount
                }
            elif type == "limit":
                body = {
                    'symbol': self._getSymbol(symbol),
                    'side': side,
                    'type': 'limit',
                    'size': amount,
                    'price': price,
                }
            body.update(params)
            apiUrl = f"/spot/orders"
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body, auth=Auth.SIGNED)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e
        
    def create_limit_order(self, symbol, amount, side, price, params={}):
        body = {
            'symbol': self._getSymbol(symbol),
            'side': side,
            'type': 'limit',
            'size': amount,
            'price': price,
        }
        body.update(params)
        apiUrl = f"/spot/v2/submit_order"
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body, auth=Auth.SIGNED)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e

    def create_market_order(self, symbol, amount, side, params={}):
        body = {
            'symbol': self._getSymbol(symbol),
            'side': side,
            'type': 'market',
            'size': amount,
            'notional': amount
        }
        body.update(params)
        apiUrl = f"/spot/v2/submit_order"
        apiUrl = f"/spot/v2/submit_order"
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body, auth=Auth.SIGNED)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        raise NotImplementedError("method not implemented")

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/spot/v4/query/open-orders"
        body = {}
        if symbol:
            body['symbol'] = self._getSymbol(symbol)
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body, auth=Auth.SIGNED)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None):
        apiUrl = "/spot/v3/cancel_order"
        body = {
            "symbol": self._getSymbol(symbol),
            "order_id": id
        }
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=body, auth=Auth.SIGNED)
            return self._parseCancelOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        raise NotImplementedError("method not implemented")
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/account/v1/wallet'
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, params=params, auth=Auth.KEYED)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
        
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        apiUrl = '/spot/v4/query/trades'
        if symbol:
            params['symbol'] = self._getSymbol(symbol)
        if limit:
            params['limit'] = limit
        if since:
            params['startTime'] = int(since)
        try:
            resp = self._signedRequest(method='POST', request_path=apiUrl, params=params, auth=Auth.SIGNED)
            return self._parseTrades(resp)
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        raise NotImplementedError("method not implemented")
    
    # parsed OHLCV = [[ts_ms, o, h ,l ,c, v],[]]
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("method not implemented")


import os, sys, json
from dotenv import load_dotenv

load_dotenv()

DIR_NAME = os.path.dirname(os.path.abspath(__file__))
root = os.path.dirname(os.path.dirname(DIR_NAME))
sys.path.append(root)

import ttxt

def getTtxtExchange(exchangeName):
    if exchangeName == "gateFutures":
        print("returned gateFuture")
        return ttxt.gateFutures(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "bybitFutures":
        print("returned bybitFutures")
        return ttxt.bybitFutures(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "bitgetFutures":
        print("returned bitgetFutures")
        return ttxt.bitgetFutures(key=os.getenv("KEY"), secret=os.getenv("SECRET"), password=os.getenv("PASSWORD"))
    if exchangeName == "bitget":
        print("returned bitget")
        return ttxt.bitget(key=os.getenv("KEY"), secret=os.getenv("SECRET"), password=os.getenv("PASSWORD"))
    if exchangeName == "kucoin":
        print("returned kucoin")
        return ttxt.kucoin(key=os.getenv("KEY"), secret=os.getenv("SECRET"), password=os.getenv("PASSWORD"))
    if exchangeName == "binance":
        print("returned binance")
        return ttxt.binance(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "bingx":
        print("returned bingx")
        return ttxt.bingx(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "biconomy":
        print("returned biconomy")
        return ttxt.biconomy(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "mexc":
        print("returned mexc")
        return ttxt.mexc(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "gate" or exchangeName == "gateio":
        print("returned gate")
        return ttxt.gateio(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "bybit":
        print("returned bybit")
        return ttxt.bybit(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "okx":
        print("returned okx")
        return ttxt.okx(key=os.getenv("KEY"), secret=os.getenv("SECRET"), password=os.getenv("PASSWORD"))
    if exchangeName == "huobi":
        print("returned huobi")
        return ttxt.huobi(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    if exchangeName == "bitmart":
        print("returned bitmart")
        return ttxt.bitmart(key=os.getenv("KEY"), secret=os.getenv("SECRET"), password=os.getenv("PASSWORD"))
    if exchangeName == "cryptocom":
        print("returned cryptocom")
        return ttxt.cryptocom(key=os.getenv("KEY"), secret=os.getenv("SECRET"))
    else:
        print(f"Exchange not supported: {exchangeName}")
        return None

def test(exchangeName="", params={}):
    ttxtExchange = getTtxtExchange(exchangeName)
    # ttxtExchange = ttxt.gateFutures(key=os.getenv("KEY"), secret=os.getenv("SECRET"))

    if "ticker" in params and params["ticker"]:
        print("testing ticker...")
        try:
            resp = ttxtExchange.fetch_ticker(tickerToTest)
            print(resp)
        except Exception as e:
            print("ticker could not be fetched")
            print(e)
    if "balance" in params and params["balance"]:
        print("testing balance...")
        try:
            resp = ttxtExchange.fetch_balance()
            print(resp)
        except Exception as e:
            print("balance could not be fetched")
            print(e)
    if "createOrder" in params and params["createOrder"]:
        print("testing create order...")
        try:
            resp = ttxtExchange.create_order(symbol=tickerToTest, order_type="limit", amount="5", side="sell", price=5)
            print(f"createOrder response: {resp}")
        except Exception as e:
            print("order could not be created")
            print(e)
    if "createLimitOrder" in params and params["createLimitOrder"]:
        print("testing limit create order...")
        try:
            resp = ttxtExchange.create_limit_order(symbol=tickerToTest, amount=50, side="buy", price=0.20)
            print(f"createLimitOrder response: {resp}")
        except Exception as e:
            print("limit order could not be created")
            print(e)
    if "createMarketSellOrder" in params and params["createMarketSellOrder"]:
        print("testing market sell create order...")
        try:
            resp = ttxtExchange.create_market_sell_order(symbol=tickerToTest, amount=0.001)
            print(f"createMarketSellOrder response: {resp}")
        except Exception as e:
            print("market sell order could not be created")
            print(e)
    if "createMarketBuyOrder" in params and params["createMarketBuyOrder"]:
        print("testing market buycreate order...")
        try:
            resp = ttxtExchange.create_market_buy_order(symbol=tickerToTest, amount=0.001)
            print(f"createMarketBuyOrder response: {resp}")
        except Exception as e:
            print("market buy order could not be created")
            print(e)
    if "create_market_order" in params and params["create_market_order"]:
        print("testing create market order...")
        try:
            resp = ttxtExchange.create_market_order(symbol=tickerToTest, side="sell", amount=1)
            print(f"create_market_order response: {resp}")
        except Exception as e:
            print("market order could not be created")
            print(e)
    if "fetchOrder" in params and params["fetchOrder"]:
        print("testing fetch order...")
        try:
            resp = ttxtExchange.fetch_order(id=45577069671, symbol=tickerToTest)
            print(f"fetchOrder response: {resp}")
        except Exception as e:
            print("order could not be fetched")
            print(e)
    if "cancelOrder" in params and params["cancelOrder"]:
        print("testing cancel order...")
        try:
            resp = ttxtExchange.cancel_order(id=698922787, symbol=tickerToTest)
            print(f"cancelOrder response: {resp}")
        except Exception as e:
            print("order could not be canceled")
            print(e)
    if "fetchOpenOrders" in params and params["fetchOpenOrders"]:
        print("testing open orders...")
        try:
            resp = ttxtExchange.fetch_open_orders(symbol=tickerToTest)
            print(f"total open orders: {len(resp)}")
            print(f"fetchOpenOrders response: {resp}")
        except Exception as e:
            print("OpenOrders could not be fetched")
            print(e)
    if "cancelAllOrders" in params and params["cancelAllOrders"]:
        print("testing open orders...")
        try:
            resp = ttxtExchange.fetch_open_orders(symbol=tickerToTest)
            print(f"total open orders: {len(resp)}")
            for order in resp:
                ttxtExchange.cancel_order(id=order['id'], symbol=tickerToTest)
            print(f"cancelAllOrders response: {resp}")
        except Exception as e:
            print("OpenOrders could not be fetched")
            print(e)
    if "setLeverage" in params and params["setLeverage"]:
        print("testing set leverage...")
        try:
            resp = ttxtExchange.set_leverage(symbol=tickerToTest, leverage=10)
            print(f"set leverage response: {resp}")
        except Exception as e:
            print("leverage could not be set")
            print(e)
    if "ohlcv" in params and params["ohlcv"]:
        print("testing ohlcv...")
        try:
            resp = ttxtExchange.fetch_ohlcv(symbol=tickerToTest, interval="hour", size=10)
            print(f"ohlcv response: {resp}")
        except Exception as e:
            print("leverage could not be set")
            print(e)
    if "orderbook" in params and params["orderbook"]:
        print("testing orderbook...")
        try:
            resp = ttxtExchange.fetch_order_book(symbol=tickerToTest, limit=100)
            print(f"orderbook response: {resp}")
        except Exception as e:
            print("orderbook could not be fetched")
            print(e)
    if "candlestick" in params and params["candlestick"]:
        print("testing candlestick...")
        try:
            resp = ttxtExchange.fetch_ohlcv(symbol=tickerToTest, timeframe='1h', limit=1)
            print(f"candlestick response: {resp}")
            print(f"candlestick response len: {len(resp)}")
        except Exception as e:
            print("candlesticks could not be fetched")
            print(e)
    if "myTrades" in params and params["myTrades"]:
        print("testing myTrades...")
        try:
            resp = ttxtExchange.fetch_my_trades(symbol=tickerToTest, since=1710863600671)
            print(resp)
        except Exception as e:
            print("myTrades could not be fetched")
            print(e)


if __name__ == "__main__":
    tickerToTest = "GALA/USDT"
    testingParams = {"ticker": False, 
                     "balance": True, 
                     "ohlcv": False, 
                     "createOrder": False, 
                     "fetchOrder": False, 
                     "cancelOrder": False, 
                     "fetchOpenOrders": False, 
                     "setLeverage": False, 
                     "createLimitOrder": False,
                     "createMarketSellOrder": False, 
                     "createMarketBuyOrder": False, 
                     "orderbook": False, 
                     "candlestick": False,
                     "create_market_order": False, 
                     "cancelAllOrders": False, 
                     "myTrades": False}
    test(params=testingParams, exchangeName="bybit")



import sys
import types
from typing import Union, List, Optional, Any
from decimal import Decimal


if sys.version_info.minor >= 8:
    from typing import TypedDict, Literal, Dict
else:
    from typing import Dict
    from typing_extensions import Literal
    TypedDict = Dict

if sys.version_info.minor >= 11:
    from typing import NotRequired
else:
    from typing_extensions import NotRequired

IndexType = Union[str, int]
Num = Union[None, str, float, int, Decimal]
Str = Optional[str]
Strings = Optional[List[str]]
Int = Optional[int]
Bool = Optional[bool]
MarketType = Literal['spot', 'margin', 'swap', 'future', 'option']
Numeric = Union[None, str, float, int, Decimal]
String = Optional[str]

class Ticker(TypedDict):
    info: Dict[str, Any]
    symbol: Str
    timestamp: Int
    datetime: Str
    high: Num
    low: Num
    bid: Num
    bidVolume: Num
    ask: Num
    askVolume: Num
    vwap: Num
    open: Num
    close: Num
    last: Num
    previousClose: Num
    change: Num
    percentage: Num
    average: Num
    quoteVolume: Num
    baseVolume: Num

class Balance(TypedDict):
    free: Num
    total: Num
    unrealisedPnl: Num
    currency: Str

class Balances(Dict[str, Balance]):
    datetime: Str
    timestamp: Int

class OrderBook(TypedDict):
    asks: List[Numeric]  # [[price, amount], []...]
    bids: List[Numeric]  # [[price, amount], []...]
    datetime: String
    timestamp: Int
    nonce: Int


from datetime import datetime, timedelta
import uuid
import time
import random

def ts_to_datetime(timestamp_ms):
    timestamp_seconds = timestamp_ms / 1000
    dateTimeObj = datetime.fromtimestamp(timestamp_seconds)
    return dateTimeObj.strftime("%Y-%m-%d %H:%M:%S")

def get_uuid():
    return str(uuid.uuid4())

def get_random_int():
    return random.randint(1, 9000000)

def get_curr_timestamp_ms():
    return int(time.time() * 1000)

def get_1d_ago_ts_ms():
    current_timestamp = datetime.now()
    timestamp_day_before = current_timestamp - timedelta(days=1)
    return int(timestamp_day_before.timestamp())

def datetime_to_ts_ms(datetime_str):
    datetime_obj = datetime.strptime(datetime_str, "%Y-%m-%d %H:%M:%S")
    return int(datetime_obj.timestamp() * 1000)


from ast import parse
from email import header
import hmac
import base64
import hashlib
import json
import time
from wsgiref import headers
import requests
from urllib.parse import urlencode
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general

class binance(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.binance.com"
        
    def _getSymbol(self, symbol):
        return symbol.replace("/", "")
    
    def generate_timestamp(self):
        """
        Return a millisecond integer timestamp.
        """
        return int(time.time() * 10**3)
    
    # Auth
    def sign(self, message):
        mac = hmac.new(self.secret.encode('utf-8'), message.encode('utf-8'), hashlib.sha256).hexdigest()
        return mac
    
    def parse_params_to_str(self, message):
        url = ''
        if message:
            url=urlencode(message)
        if url == '':
            return ''
        return url
    
    def _signedRequest(self, method, request_path, queryString, body):
        timeMs = self.generate_timestamp()
        if method == "POST":
            body["timestamp"] = timeMs
            bodyParse = self.parse_params_to_str(body)
            query = self.parse_params_to_str(queryString)
            body['signature'] = self.sign(queryString+bodyParse)
            body = query+self.parse_params_to_str(body)
            headers = {
                "X-MBX-APIKEY": self.key
            }
            if query == '':
                url = self.domain_url+request_path
            else:
                url = self.domain_url+request_path+'?'+query
            try:
                response = requests.request(method, url, headers=headers, data=body)
                return response.json()
            except Exception as e:
                raise e
        elif method == "GET"  or method == "DELETE":
            body = json.dumps(body)
            queryString["timestamp"] = timeMs
            queryString['signature'] = self.sign(self.parse_params_to_str(queryString))
            query = self.parse_params_to_str(queryString)
            headers = {
                "X-MBX-APIKEY": self.key
            }
            url = self.domain_url+request_path+'?'+query
            try:
                response = requests.request(method, url, headers=headers)
                return response.json()
            except Exception as e:
                raise e
            
    # Parsers
    def _parseBalance(self, balData):
        parsedBal = {"free": {}, "total": {}}
        if 'code' in balData:
            raise Exception(balData['msg'])
        data = balData.get("balances", None)
        for asset in data:
            if float(asset['free']) or float(asset['locked']):
                parsedBal["free"][asset["asset"]] = asset.get("free", None) 
                parsedBal["total"][asset["asset"]] = str(float(asset.get("free", None))+float(asset.get("locked", None)))
        return parsedBal
    
    def _parseCreateorder(self, order):
        parsedOrder = {'id': None, "symbol": None, "amount": None, "side": None, "timestamp": None, "status": None, "orderJson": None}
        if 'code' in order:
            raise Exception(order['msg'])
        parsedOrder['id'] = order.get("orderId", None)
        parsedOrder['symbol'] = order.get("symbol", None)
        parsedOrder['amount'] = order.get("executedQty", None)
        parsedOrder['side'] = order.get("side", None)
        parsedOrder['status'] = order.get("type", None)
        parsedOrder['timestamp'] = order.get("transactTime", None)
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder
    
    def _parseOpenOrders(self, orders):
        if 'code' in orders:
            raise Exception(orders['msg'])
        parsedOrderList = []
        for order in orders:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('orderId', None)
        parsedOrder["symbol"] = order.get('symbol', None)
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["origQty"]) if "origQty" in order else None
        parsedOrder["side"] = order['side'].lower() if "side" in order else None
        parsedOrder["timestamp"] = order["updateTime"] if "updateTime" in order else None
        parsedOrder["status"] = order.get("status", None)
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseCancelorder(self, order):
        parsedOrder = {'id': None, "symbol": None, "amount": None, "side": None, "timestamp": None, "status": None, "orderJson": None}
        if 'code' in order:
            raise Exception(order['msg'])
        parsedOrder['id'] = order.get("orderId", None)
        parsedOrder['symbol'] = order.get("symbol", None)
        parsedOrder['amount'] = order.get("origQty", None)
        parsedOrder['side'] = order.get("side", None)
        parsedOrder['status'] = order.get("status", None)
        if order.get("transactTime"):
            parsedOrder['timestamp'] = int(order.get("transactTime"))
        else:
            parsedOrder['timestamp'] = None
        return parsedOrder
    
    def _parseFetchTrades(self, orderData):
        parsedTradesList = []
        if 'code' in orderData:
            raise Exception(orderData['msg'])
        for order in orderData:
            parsedTradesList.append(self._parseTrades(order))
        return parsedTradesList
    
    def _parseTrades(self, order):
        parsedOrder = {"id": None, "symbol": None, "price": None, "amount": None, "takerOrMaker": None, "datetime": None, "fee": None, "fee_currency": None,
                       "side": None, "timestamp": None, "status": None}
        parsedOrder['id'] = order.get("orderId", None)
        parsedOrder["tradeId"] = order.get("id", None)
        parsedOrder['symbol'] = order.get("symbol", None)
        parsedOrder['amount'] = order.get("qty", None)
        parsedOrder['price'] = order.get("price", None)
        parsedOrder["side"] = "buy" if order["isBuyer"] else "sell"
        if order["isMaker"]:
            parsedOrder['takerOrMaker'] = 'maker'
        else:
            parsedOrder['takerOrMaker'] = 'taker'
        if order.get("time"):
            parsedOrder['timestamp'] = int(order.get("time"))
        else:
            parsedOrder['timestamp'] = None
        parsedOrder['datetime'] = general.ts_to_datetime(parsedOrder["timestamp"]) if parsedOrder['timestamp'] else None
        parsedOrder["status"] = order.get('status', None)
        parsedOrder["fee_currency"] = order.get('commissionAsset', None)
        parsedOrder['fee'] = order.get("commission", None)
        return parsedOrder
    
    # Exchange functions 
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = "/api/v3/account"
        try:
            params = {
            }
            response = self._signedRequest('GET', apiUrl, queryString=params, body='')
            return self._parseBalance(response)
        except Exception as e:
            raise e
    
    def create_order(self, symbol, side, amount, order_type, price=None, params={}):
        apiUrl = "/api/v3/order"
        try:
            params = {
                "symbol": self._getSymbol(symbol),
                "side": side.upper(),
                "type": order_type.upper(),
                "quantity": amount
            }
            if order_type == "limit":
                params["price"] = price
                params["timeInForce"] = "GTC"
            response = self._signedRequest('POST', request_path=apiUrl, body=params, queryString='')
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def create_market_order(self, symbol, side, amount, params={}):
        return self.create_order(symbol, side, amount=amount, price=None, order_type='market', params={})
    
    def create_limit_order(self, symbol, side, amount, price, params={}):
        return self.create_order(symbol, side, amount, order_type='limit', price=price, params={})
    
    def cancel_order(self, id, symbol=None, params={}):
        apiUrl = "/api/v3/order"
        try:
            params = {
                "orderId": id
            }
            if symbol:
                params["symbol"] = self._getSymbol(symbol)
            response = self._signedRequest('DELETE', request_path=apiUrl, queryString=params, body='')
            return self._parseCancelorder(response)
        except Exception as e:
            raise e
    
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}): 
        apiUrl = "/api/v3/myTrades"
        try:
            params = {
                "symbol": self._getSymbol(symbol),
            }
            if since:
                params["startTime"] = since
            if 'endTime' in params and params['endTime']:
                params['endAt'] = params['endTime']
            if limit:
                params['limit'] = limit
            response = self._signedRequest('GET', request_path=apiUrl, queryString=params, body='')
            return self._parseFetchTrades(response)
        except Exception as e:
            raise e
    
    def fetch_open_orders(self, symbol=None, kwargs=None):
        apiUrl = "/api/v3/openOrders"
        try:
            params = {
                "symbol": self._getSymbol(symbol)
            }
            response = self._signedRequest("GET", request_path=apiUrl, queryString=params, body='')
            return self._parseOpenOrders(response)
        except Exception as e:
            raise e


import hmac
import base64
import hashlib
import json
import time
import requests
from datetime import datetime as dt
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general

'''
kwargs = {}
'''
class mexc(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.mexc.com"
        self.max_retries = 5
        self.symbolLotSize = {}
        self.recvWindow = 60000

    def _getSymbol(self, symbol):
        return symbol.replace("/", "")

    ## Auth 
    def get_expire(self):
        return int((time.time() + 1) * 1000)  # websockets use seconds
    
    def generate_timestamp(self):
        """
        Return a millisecond integer timestamp.
        """
        return int(time.time() * 10**3)
    
    def _get_sign(self, payload):
        # payload = ":"+ payload + "&secret_key=" + self.secret
        payload = payload + "&secret_key=" + self.secret
        signature = hashlib.md5(payload.encode()).hexdigest().upper()
        return signature
        
    def _parseParams(self, paramsMap):
        sortedKeys = sorted(paramsMap)
        paramsStr = "&".join(["%s=%s" % (x, paramsMap[x]) for x in sortedKeys])
        # return paramsStr+"&timestamp="+str(int(time.time() * 1000))
        return paramsStr
    
    def _check_resp_status(self, resp):
        if "code" in resp:
            if resp["code"] != 0:
                raise Exception(resp["msg"])
            else:
                return resp  # success 
        if "message" in resp:
            raise Exception(resp["message"])
        return resp
    
    def _authenticate(self,params):
        params = self._prepare_params(params)
        signature = hmac.new(
            self.secret.encode('utf-8'), 
            params.encode('utf-8'),
            hashlib.sha256
        ).hexdigest() 
        headers = {
            'X-MEXC-APIKEY': self.key
        }
        updatedParams = params + f"&signature={signature}"
        return headers, updatedParams
        
    def _send_request(self, method, path, params, payload={}):
        headers, updatedParams = self._authenticate(params)
        try:
            updatedurl = self.domain_url + path + "?"+ updatedParams
            headers.update({'Content-Type': 'application/json'})
            if method == "GET":
                response = requests.get(updatedurl, headers=headers)
                return self._check_resp_status(response.json())
            if method == "POST":
                response = requests.post(updatedurl, headers=headers)
                return self._check_resp_status(response.json())
            if method == "DELETE":
                response = requests.delete(updatedurl, headers=headers)
                return self._check_resp_status(response.json())
        except Exception as e:
            raise e
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")
        
    def _prepare_params(self, params):
        payload = "&".join(
                [
                    str(k) + "=" + str(v)
                    for k, v in params.items()  #use for sort: sorted(params.items())
                    if v is not None
                ]
            )
        return payload

    ## parsers
    def _parseCreateorder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('orderId', None)
        parsedOrder["symbol"] = order.get('symbol', None)
        parsedOrder["price"] = order.get("price", None)
        parsedOrder["amount"] = order.get("origQty", None)
        if "side" in order: 
            parsedOrder["side"] = order["side"].lower()
        else:
            parsedOrder["side"] = None
        parsedOrder["timestamp"] = order.get("transactTime", None)
        parsedOrder["status"] = "closed"
        parsedOrder["orderJson"] = json.dumps(order)
        return parsedOrder

    '''
    {'makerCommission': None, 'takerCommission': None, 'buyerCommission': None, 
    'sellerCommission': None, 'canTrade': True, 'canWithdraw': True, 'canDeposit': True, 
    'updateTime': None, 'accountType': 'SPOT', 'balances': [{'asset': 'USDT', 
    'free': '59727.7100003983', 'locked': '0'}, {'asset': 'USDC', 'free': '2216', 
    'locked': '0'}, {'asset': 'GAL', 'free': '89.65', 'locked': '0'}, {'asset': 
    'OPUL', 'free': '790.15', 'locked': '0'}], 'permissions': ['SPOT']}
    '''
    def _parseBalance(self, balanceData):
        parsedBal = {"free": {}, "total": {}}
        if "balances" not in balanceData:
            return parsedBal
        for bal in balanceData["balances"]:
            parsedBal["free"][bal["asset"]] = bal.get("free", None)
            if "free" in bal and "locked" in bal:
                parsedBal["total"][bal["asset"]] = float(bal["free"]) + float(bal["locked"])
            else: parsedBal["free"][bal["asset"]] = bal.get("free", None)
        return parsedBal

    def _parseOpenOrders(self, orders):
        parsedOrderList = []
        for order in orders:
            parsedOrderList.append(self._parseCreateorder(order))
        return parsedOrderList
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("timestamp", None)
        parsedData["nonce"] = orderbookData.get("id", None)
        if "bids" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["bids"]]
        else: parsedData["bids"] = []
        if "asks" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["asks"]]
        else: parsedData["asks"] = []
        return parsedData

    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["symbol"] = trade.get("symbol", None)
            parsedTrade["id"] = trade.get("orderId", None)
            parsedTrade["tradeId"] = trade.get("id", None)
            parsedTrade["side"] = "buy" if trade["isBuyer"] else "sell"
            parsedTrade["price"] = trade.get("price", None)
            parsedTrade["amount"] = trade.get("qty", None)
            parsedTrade["takerOrMaker"] = "maker" if trade["isMaker"] else "taker"
            if "time" in trade:
                parsedTrade["timestamp"] = int(float(trade["time"]))
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            parsedTrade["feeCurrency"] = trade["commissionAsset"]
            parsedTrade["fee"] = float(trade["commission"])
            return parsedTrade
        parsedTradeList = []
        for trade in trades:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList

    ## Exchange functions
    # buy orders size are in quote and sell orders are in base
    def create_market_order(self, symbol, side, amount, params={}, price=None):
        apiUrl = "/api/v3/order"
        try:
            ticker = self._getSymbol(symbol)
            params = {
                'symbol': self._getSymbol(ticker),
                'side': side.upper(),
                'type': 'MARKET',
                'quantity': float(amount),
                'quoteOrderQty': float(amount),
                'recvWindow': self.recvWindow,
                'timestamp': self.generate_timestamp()
            }
            params.update(params) 
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def create_limit_order(self, symbol, side, amount, price, params={}):
        apiUrl = "/api/v3/order"
        try:
            ticker = self._getSymbol(symbol)
            params = {
                'symbol': self._getSymbol(ticker),
                'side': side.upper(),
                'type': 'LIMIT',
                'quantity': float(amount),
                'price': float(price),
                'quoteOrderQty': float(amount),
                'recvWindow': self.recvWindow,
                'timestamp': self.generate_timestamp()
            }
            params.update(params) 
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        raise NotImplementedError("method not implemented")
    
    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = '/api/v3/openOrders'
        params = {
            'symbol': self._getSymbol(symbol),
            'recvWindow': self.recvWindow,
            'timestamp': self.generate_timestamp()
        }
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=params)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None):
        apiUrl = '/api/v3/order'
        params = {
            'symbol': self._getSymbol(symbol),
            'orderId': id,
            'recvWindow': self.recvWindow,
            'timestamp': self.generate_timestamp()
        }
        try:
            resp = self._send_request(method='DELETE', path=apiUrl, params=params)
            return self._parseCreateorder(resp)
        except Exception as e:
            raise e
        
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/api/v3/account'
        params = {
            'recvWindow': self.recvWindow,
            'timestamp': self.generate_timestamp()
        }
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=params)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
    
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        apiUrl = '/api/v3/myTrades'
        params = {
            'timestamp': self.generate_timestamp(),
            'recvWindow': self.recvWindow
        }
        if symbol:
            params['symbol'] = self._getSymbol(symbol)
        if limit:
            params['limit'] = limit
        if since:
            params['startTime'] = int(since)
        if "endTime" in params:
            params['endTime'] = int(params["endTime"])
        try:
            resp = self._send_request(method='GET', path=apiUrl, params=params)
            return self._parseTrades(resp)
        except Exception as e:
            raise e
    
    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        raise NotImplementedError("method not implemented")

    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        apiUrl = f"/api/v3/depth"
        queryParams = {"symbol": self._getSymbol(symbol)}
        if limit is not None:
            queryParams["limit"] = int(limit)
        queryParams.update(params)
        try:
            resp = self._unsignedRequest('GET', apiUrl, params=queryParams)
            return self._parseOrderbook(resp)
        except Exception as e:
            raise e


import hmac
import base64
import hashlib
import json
import time
import requests
from urllib.parse import urlencode
from ttxt.base import baseFuturesExchange
from ttxt.types import baseTypes
# Example of an exchange requiring a password
'''
Sample 
kwargs = {
    "productType": "",
    "marginMode": "",
    "marginCoin": ""
}
'''
class bitgetFutures(baseFuturesExchange.BaseFuturesExchange):
    def __init__(self, key, secret, password, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.password = password
        self.domain_url = "https://api.bitget.com"
        self.productType =  kwargs["productType"] if "productType" in kwargs else "USDT-FUTURES"
        self.marginMode = kwargs["marginMode"] if "marginMode" in kwargs else "isolated"
        self.marginCoin = kwargs["marginCoin"] if "marginCoin" in kwargs else "USDT"
        self.success_sode = '00000'

    def sign(self, message):
        mac = hmac.new(bytes(self.secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return base64.b64encode(d)

    def pre_hash(self, timestamp, method, request_path, queryString, body=None):
        if not body:
            return str(timestamp) + str.upper(method) + request_path + queryString
        return str(timestamp) + str.upper(method) + request_path + body

    def parse_params_to_str(self, params):
        params = [(key, val) for key, val in params.items()]
        params.sort(key=lambda x: x[0])
        url = '?' + urlencode(params)
        if url == '?':
            return ''
        return url

    def _signedRequest(self, method, request_path, queryString, body):
        timeMs = int(time.time()) *1000
        if method == "POST":
            body = json.dumps(body)
            queryString = json.dumps(queryString) if queryString is not None else ''
            signature = self.sign(self.pre_hash(timeMs, method, request_path, queryString ,body))
            headers = {"ACCESS-KEY": self.key, "ACCESS-SIGN": signature, "ACCESS-PASSPHRASE": self.password, "ACCESS-TIMESTAMP": str(timeMs), "locale": "en-US", "Content-Type": "application/json"}
            url = self.domain_url+request_path
            try:
                response = requests.post(url, headers=headers, data=body)
                return response.json()
            except Exception as e:
                raise e
        if method == "GET":
            try:
                params = body
                body = ""
                request_path = request_path + self.parse_params_to_str(params) # Need to be sorted in ascending alphabetical order by key
                signature = self.sign(self.pre_hash(timeMs, "GET", request_path, str(body)))

                headers = {"ACCESS-KEY": self.key, "ACCESS-SIGN": signature, "ACCESS-PASSPHRASE": self.password, "ACCESS-TIMESTAMP": str(timeMs), "locale": "en-US", "Content-Type": "application/json"}
                url = self.domain_url+request_path
                response = requests.get(url, headers=headers)
                return response.json()
            except Exception as e:
                raise e
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")


    ## parsers
    def _parseBalance(self, balData):
        parsedBal = {}
        parsedBal["free"] = balData.get("available", None)
        parsedBal["total"] = balData.get("total", None)
        parsedBal["unrealisedPnl"] = balData.get("unrealised_pnl", None)
        parsedBal["currency"] = balData.get("currency", None)
        return parsedBal


    ## Exchange functions 
    def fetch_ticker(self, symbol):
        print(f"Fetching ticker for {symbol} with password {self.password}")
        
    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            if type == "market":
                body = {
                    "symbol": symbol,
                    "productType": self.productType,
                    "marginMode": self.marginMode,
                    "marginCoin":self.marginCoin,
                    "size": amount,
                    "side": side,
                    "tradeSide": params["tradeSide"] if "tradeSide" in params else "",
                    "orderType": "market",
                    "force": params["force"] if "force" in params else "",
                    "reduceOnly": params["reduceOnly"] if "reduceOnly" in params else "NO",
                    "presetStopSurplusPrice": params["tpPrice"] if "tpPrice" in params else "",
                    "presetStopLossPrice": params["slPrice"] if "slPrice" in params else "",
                }
            elif type == "limit":
                body = {
                    "symbol": symbol,
                    "productType": self.productType,
                    "marginMode": self.marginMode,
                    "marginCoin":self.marginCoin,
                    "size": amount,
                    "price": price,
                    "side": side,
                    "tradeSide": params["tradeSide"] if "tradeSide" in params else "",
                    "orderType": "market",
                    "force": params["force"] if "force" in params else "",
                    "reduceOnly": params["reduceOnly"] if "reduceOnly" in params else "NO",
                    "presetStopSurplusPrice": params["tpPrice"] if "tpPrice" in params else "",
                    "presetStopLossPrice": params["slPrice"] if "slPrice" in params else "",
                }
            # body = '{"symbol":' + '\"' + demoSymbol + '",' + '\"marginCoin":"SUSDT","side":"open_long","orderType":"market","size":' + '\"' + str(quantity) + '\"}'
            apiUrl = "/api/v2/mix/order/place-order"
            response = self._signedRequest('POST', apiUrl, None, body)
            return response
        except Exception as e:
            raise e
        
    # def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
    #     apiUrl = "/api/v2/mix/market/ticker"
    #     params = {
    #         "symbol": symbol,
    #         "productType": self.productType
    #     }
    #     resp = self._unsignedRequest('GET', apiUrl, params=params)
    #     return resp # parse this response into Ticker
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        pass

    def fetch_account(self, symbol, params={}):
        try:
            body = {
                "symbol": symbol,
                "productType": self.productType,
                "marginCoin":self.marginCoin,
            }
            apiUrl = "/api/v2/mix/account/account"
            response = self._signedRequest('GET', apiUrl, None, body)
            if response['code'] == self.success_sode:
                return response['data']
            return response
        except Exception as e:
            raise e
        
    def fetch_accounts(self, params={}):
        try:
            body = {
                "productType": self.productType,
            }
            apiUrl = "/api/v2/mix/account/accounts"
            response = self._signedRequest('GET', apiUrl, None, body)
            if response['code'] == self.success_sode:
                return response['data']
            return response
        except Exception as e:
            raise e
        
    def fetch_position(self, symbol, params={}):
        try:
            body = {
                "symbol": symbol,
                "productType": self.productType,
                "marginCoin":self.marginCoin,
            }
            apiUrl = "/api/v2/mix/position/single-position"
            response = self._signedRequest('GET', apiUrl, None, body)
            if response['code'] == self.success_sode:
                return response['data']
            return response
        except Exception as e:
            raise e

    def fetch_positions(self, params={}):
        try:
            body = {
                "productType": self.productType,
                "marginCoin": self.marginCoin,
            }
            apiUrl = "/api/v2/mix/position/all-position"
            response = self._signedRequest('GET', apiUrl, None, body)
            if response['code'] == self.success_sode:
                return response['data']
            return response
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol):
        apiUrl = "/api/v2/mix/market/ticker"
        params = {
            "symbol": symbol,
            "productType": self.productType
        }
        resp = self._unsignedRequest('GET', apiUrl, params=params)
        return resp

    def set_leverage(self, symbol, leverage):
        apiUrl = "/api/v2/mix/account/set-leverage"
        params = {
            "symbol": symbol,
            "productType": self.productType,
            "marginCoin": self.marginCoin,
            "leverage": leverage,
            "holdSide": "long"
        }
        response_long = self._signedRequest('POST', apiUrl, None, params)
        params["holdSide"] = "short"
        response_short = self._signedRequest('POST', apiUrl, None, params)
        return True

    def cancel_order(self, id, params={}):
        pass

    def fetch_ohlcv(self, symbol, timeframe, startTime, endTime=None, limit=200):
        params = {
            'symbol': symbol,
            'granularity': timeframe,
            'startTime': startTime,
            'endTime': endTime,
            'limit': limit,
            'productType': 'usdt-futures'
        }
        if endTime is None and limit is None:
            print("endTime or limit must be specified")
            return []
        if endTime is None:
            params.pop('endTime')
        if limit is None:
            params.pop('limit')

        full_url = f"https://api.bitget.com/api/v2/mix/market/history-candles?"+'&'.join([f'{k}={v}' for k, v in params.items()])
        retry = True
        while retry:
            response = requests.get(full_url)
            if response.status_code == 200:
                candle_data = json.loads(response.content)['data']
                if len(candle_data) > 0:
                    return candle_data
            else:
                break
        return []


import hmac
import base64
import hashlib
import json
import time
import requests
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general


class gateio(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.gateio.ws"  # host + prefix
        self.prefix = "/api/v4"
        self.settle_currency = "usdt"
        self.quanto_multiplier = {}

    def _getSymbol(self, symbol):
        return symbol.replace("/", "_")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("_", "/")

    ## Auth 
    def gen_sign(self, method, url, query_string=None, payload_string=None):
        key = self.key
        secret = self.secret
        t = time.time()
        m = hashlib.sha512()
        m.update((payload_string or "").encode('utf-8'))
        hashed_payload = m.hexdigest()
        s = '%s\n%s\n%s\n%s\n%s' % (method, url, query_string or "", hashed_payload, t)
        sign = hmac.new(secret.encode('utf-8'), s.encode('utf-8'), hashlib.sha512).hexdigest()
        return {'KEY': key, 'Timestamp': str(t), 'SIGN': sign}

    def connectToPrivateWss(self, creds, ticker):
        request = {
            "time": int(time.time()),
            "channel": "spot.orders",
            "event": "subscribe",  # "unsubscribe" for unsubscription
            "payload": [ticker]
        }
        # refer to Authentication section for gen_sign implementation
        request['auth'] = self.gen_sign(request['channel'], request['event'], request['time'], creds)
        return json.dumps(request) 

    ## Requests 
    def _signedRequest(self, request_path, method, queryString, body):
        common_headers = {'Accept': 'application/json', 'Content-Type': 'application/json'}
        if method == "GET":
            try:
                if queryString != "" and len(queryString):
                    url = self.domain_url+ self.prefix + request_path + "?" + queryString
                else:
                    url = self.domain_url+ self.prefix + request_path
                sign_headers = self.gen_sign('GET', self.prefix + request_path, queryString, "")
                sign_headers.update(common_headers)
                response = requests.get(url, headers=sign_headers)
                return response.json()
            except Exception as e:
                raise e
        if method == "POST":
            try:
                if queryString and queryString != "" and len(queryString):
                    url = self.domain_url+ self.prefix + request_path + "?" + queryString
                    sign_headers = self.gen_sign('POST', self.prefix + request_path, queryString, "")
                    sign_headers.update(common_headers)
                    response = requests.post(url, headers=sign_headers)
                else:
                    url = self.domain_url+ self.prefix + request_path
                    request_content = json.dumps(body)
                    sign_headers = self.gen_sign('POST', self.prefix + request_path, "", request_content)
                    sign_headers.update(common_headers)
                    url = self.domain_url + self.prefix + request_path
                    response = requests.post(url, headers=sign_headers, data=request_content)
                return response.json()
            except Exception as e:
                raise e
        if method == "DELETE":
            try:
                if queryString != "" and len(queryString):
                    url = self.domain_url+ self.prefix + request_path + "?" + queryString
                else:
                    url = self.domain_url+ self.prefix + request_path
                sign_headers = self.gen_sign('DELETE', self.prefix + request_path, queryString, "")
                sign_headers.update(common_headers)
                response = requests.delete(url, headers=sign_headers)
                return response.json()
            except Exception as e:
                raise e
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + self.prefix + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")

    ## parsers
    # gateio doesnt return 'total' for spot accounts so, total=available
    def _parseBalance(self, balData):
        if 'label' in balData: raise Exception(balData['message'])
        parsedBal = {"free": {}, "total": {}}
        for bal in balData:
            parsedBal["free"][bal["currency"]] = bal.get("available", None)
            parsedBal["total"][bal["currency"]] = bal.get("total", parsedBal["free"][bal["currency"]])
        return parsedBal
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("current", None)
        parsedData["nonce"] = orderbookData.get("id", None)
        if "bids" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["bids"]]
        else: parsedData["bids"] = []
        if "asks" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["asks"]]
        else: parsedData["asks"] = []
        return parsedData
    
    '''
    {'id': '523236767452', 'text': 'apiv4', 'amend_text': '-', 'create_time': '1709553380', 
    'update_time': '1709553380', 'create_time_ms': 1709553380379, 'update_time_ms': 1709553380379,
      'status': 'open', 'currency_pair': 'GAL_USDT', 'type': 'limit', 'account': 'spot', 
      'side': 'buy', 'amount': '1', 'price': '3.3', 'time_in_force': 'gtc', 'iceberg': '0', 
      'left': '1', 'filled_amount': '0', 'fill_price': '0', 'filled_total': '0', 'fee': '0', 
      'fee_currency': 'GAL', 'point_fee': '0', 'gt_fee': '0', 'gt_maker_fee': '0', 
      'gt_taker_fee': '0', 'gt_discount': False, 'rebated_fee': '0', 
      'rebated_fee_currency': 'USDT', 'finish_as': 'open'}
    '''
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('id', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["currency_pair"]) if "currency_pair" in order else None
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["amount"]) if "amount" in order else None
        parsedOrder["side"] = order.get('side', None)
        parsedOrder["timestamp"] = int(float(order["create_time"])) if "create_time" in order else None
        parsedOrder["status"] = order["status"] if "status" in order else None
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseCreateOrder(self, order):
        if 'label' in order: raise Exception(order['message'])
        parsedOrder = {}
        parsedOrder["id"] = order.get('id', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["currency_pair"]) if "currency_pair" in order else None
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["amount"]) if "amount" in order else None
        parsedOrder["side"] = order.get('side', None)
        parsedOrder["timestamp"] = int(float(order["create_time"])) if "create_time" in order else None
        parsedOrder["status"] = order["status"] if "status" in order else None
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseOpenOrders(self, orders):
        parsedOrderList = []
        if 'label' in orders: raise Exception(orders['message'])
        for order in orders:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    '''
    {'id': '7697129236', 'create_time': '1709562413', 'create_time_ms': '1709562413816.639000', 
    'currency_pair': 'GAL_USDT', 'side': 'buy', 'role': 'taker', 'amount': '0.29', 
    'price': '3.3558', 'order_id': '523332716519', 'fee': '0.0002465', 
    'fee_currency': 'GAL', 'point_fee': '0.0', 'gt_fee': '0.0', 'amend_text': '-', 
    'sequence_id': '380802', 'text': 'apiv4'}
    '''
    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["id"] = trade.get("order_id", None)
            parsedTrade["tradeId"] = trade.get("id", None)
            parsedTrade["side"] = trade.get("side", None)
            parsedTrade["price"] = trade.get("price", None)
            parsedTrade["amount"] = trade.get("amount", None)
            parsedTrade["takerOrMaker"] = trade.get("role", None)
            if "create_time_ms" in trade:
                parsedTrade["timestamp"] = int(float(trade["create_time_ms"]))
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            return parsedTrade
        parsedTradeList = []
        if 'label' in trades: raise Exception(trades['message'])
        for trade in trades:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList

    ## Exchange functions 

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            if type == "market":
                body = {
                    "currency_pair": self._getSymbol(symbol),
                    "type": "market",
                    "account": "spot",
                    "side": side,
                    "iceberg": params["iceberg"] if "iceberg" in params else "0",
                    "amount": str(amount),
                    "time_in_force": params["time_in_force"] if "time_in_force" in params else "ioc",
                }
            elif type == "limit":
                body = {
                    "currency_pair": self._getSymbol(symbol),
                    "type": "limit",
                    "account": "spot",
                    "side": side,
                    "iceberg": params["iceberg"] if "iceberg" in params else "0",
                    "amount": str(amount),
                    "price": str(price),
                    "time_in_force": params["time_in_force"] if "time_in_force" in params else "gtc",
                    "auto_borrow": params["auto_borrow"] if "auto_borrow" in params else False,
                }
            apiUrl = f"/spot/orders"
            response = self._signedRequest(method='POST', request_path=apiUrl, queryString="", body=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e
        
    def create_limit_order(self, symbol, amount, side, price, params={}):
        body = {
            "currency_pair": self._getSymbol(symbol),
            "type": "limit",
            "account": "spot",
            "side": side,
            "iceberg": params["iceberg"] if "iceberg" in params else "0",
            "amount": str(amount),
            "price": str(price),
            "time_in_force": params["time_in_force"] if "time_in_force" in params else "gtc",
            "auto_borrow": params["auto_borrow"] if "auto_borrow" in params else False,
        }
        apiUrl = f"/spot/orders"
        try:
            response = self._signedRequest(method='POST', request_path=apiUrl, queryString="", body=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e

    def create_market_order(self, symbol, amount, side, params={}):
        body = {
            "currency_pair": self._getSymbol(symbol),
            "type": "market",
            "account": "spot",
            "side": side,
            "iceberg": params["iceberg"] if "iceberg" in params else "0",
            "amount": str(amount),
            "time_in_force": params["time_in_force"] if "time_in_force" in params else "ioc",
        }
        apiUrl = f"/spot/orders"
        try:
            response = self._signedRequest(method='POST', request_path=apiUrl, queryString="", body=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        raise NotImplementedError("method not implemented")

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/spot/orders"
        try:
            query_param = f'currency_pair={self._getSymbol(symbol)}&status=open'
            resp = self._signedRequest(method='GET', request_path=apiUrl, queryString=query_param, body=None)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None):
        apiUrl = f"/spot/orders/{id}"
        query_param = f'currency_pair={self._getSymbol(symbol)}'
        try:
            resp = self._signedRequest(method='DELETE', request_path=apiUrl, queryString=query_param, body=None)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        raise NotImplementedError("method not implemented")
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/spot/accounts'
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, body=None, queryString=params)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
        
    def fetch_my_trades(self, symbol, since=None, limit=None, params={}):
        apiUrl = '/spot/my_trades'
        query_params = ''
        if symbol:
            query_params += f'currency_pair={self._getSymbol(symbol)}'
        if limit:
            query_params += f'&limit={limit}'
        if since:
            query_params += f'&from={int(since)}'
        if 'to' in params and params['to']:
            query_params += f'&to={int(params["to"])}'
        if 'page' in params and params['page']:
            query_params += f'&page={int(params["page"])}'
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, body=None, queryString=query_params)
            return self._parseTrades(resp)
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        apiUrl = f"/spot/order_book"
        queryParams = {"currency_pair": self._getSymbol(symbol)}
        if limit is not None:
            queryParams["limit"] = int(limit)
        queryParams.update(params)
        try:
            resp = self._unsignedRequest('GET', apiUrl, params=queryParams)
            return self._parseOrderbook(resp)
        except Exception as e:
            raise e
    
    # parsed OHLCV = [[ts_ms, o, h ,l ,c, v],[]]
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("method not implemented")


import hmac
import base64
import hashlib
import json
import time
import requests
from datetime import datetime as dt
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes

'''
kwards = {}
'''
class biconomy(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://www.biconomy.com"
        self.max_retries = 5
        self.symbolLotSize = {}

    def _getSymbol(self, symbol):
        return symbol.replace("/", "_")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("_", "/")
    
    ## Auth 
    def get_expire(self):
        return int((time.time() + 1) * 1000)  # websockets use seconds
    
    def generate_timestamp(self):
        """
        Return a millisecond integer timestamp.
        """
        return int(time.time() * 10**3)
    
    def _get_sign(self, payload):
        # payload = ":"+ payload + "&secret_key=" + self.secret
        payload = payload + "&secret_key=" + self.secret
        signature = hashlib.md5(payload.encode()).hexdigest().upper()
        return signature
        
    def _parseParams(self, paramsMap):
        sortedKeys = sorted(paramsMap)
        paramsStr = "&".join(["%s=%s" % (x, paramsMap[x]) for x in sortedKeys])
        # return paramsStr+"&timestamp="+str(int(time.time() * 1000))
        return paramsStr
    
    def _check_resp_status(self, resp):
        if "code" in resp:
            if resp["code"] != 0:
                raise Exception(resp["message"])
            else:
                return resp  # success 
        if "message" in resp:
            raise Exception(resp["message"])
        raise Exception("Unknown error")

    def _send_request(self, method, path, params, payload={}):
        # self._send_request(method='POST', path=apiUrl, params=params, payload={})
        # url = "%s%s?%s&sign=%s" % (self.domain_url, path, params, self._get_sign(params))
        paramsForReq = params.copy()
        params.update({"api_key": self.key})
        params = self._parseParams(params)
        sign = self._get_sign(params)
        # params = params + "&sign=" + sign
        url = "%s%s" % (self.domain_url, path)
        headers = {
            'X-SITE-ID': "127",
            "Content-type": "application/x-www-form-urlencoded"
        }
        try:
            updatedurl = url + "?api_key=" + self.key + "&sign=" + sign
            response = requests.post(updatedurl, headers=headers, data=paramsForReq)
            return self._check_resp_status(response.json())
        except Exception as e:
            raise e
        
        
    def _prepare_params(self, params):
        payload = "&".join(
                [
                    str(k) + "=" + str(v)
                    for k, v in sorted(params.items())
                    if v is not None
                ]
            )
        return payload

    def _unsignedRequest(self, method=None, path=None, query=None, auth=False):
        path = self.domain_url + path
        if query is None:
            query = {}
        # Bug fix: change floating whole numbers to integers to prevent
        # auth signature errors.
        if query is not None:
            for i in query.keys():
                if isinstance(query[i], float) and query[i] == int(query[i]):
                    query[i] = int(query[i])
        # Send request and return headers with body. Retry if failed.
        retries_attempted = self.max_retries
        req_params = None
        while True:
            retries_attempted -= 1
            if retries_attempted < 0:
                raise Exception(
                    "Bad Request. Retries exceeded maximum."
                )
            # req_params = self._prepare_params(query)
            req_params = self._prepare_params(query)
            # Authenticate if we are using a private endpoint.
            headers = {'X-SITE-ID': "127"}
            if method == "GET":
                try:
                    if req_params:
                        client = requests.Session()
                        r = client.prepare_request(requests.Request(method, path, headers=headers))
                        r = requests.get(path + f"?{req_params}", headers=headers)
                    else:
                        r = requests.get(path, headers=headers)
                    return r.json()
                except Exception as e:
                    raise e
            if method == "POST":
                r = requests.post( path, data=req_params, headers=headers)
                return r.json()
            if method == "DELETE":
                r = requests.delete( path, data=req_params, headers=headers)
                return r.json()

    ## parsers
    def _parseBalance(self, balData):
        parsedBal= {"free": {}, "total": {}}
        if "result" in balData and balData["result"]:
            for key, val in balData["result"].items():
                if key != "user_id":
                    parsedBal["free"][key] = float(val["available"])
                    parsedBal["total"][key] = float(val["available"]) + float(val["freeze"])
        return parsedBal
    
    '''
    {'code': 0, 'message': 'Successful operation', 'result': {'amount': '0.001', 'ctime': 1703112715.647542, 
    'deal_fee': '0', 'deal_money': '0', 'deal_stock': '0', 'id': 45576367001, 'left': '0.001', 'maker_fee': '0',
    'market': 'BTC_USDT', 'mtime': 1703112715.647542, 'price': '41900', 'side': 2, 'source': 'api,127', 'taker_fee': '0', 
    'type': 1, 'user': 51340387}}
    '''
    def _parseCreateorder(self, order, openOrders=False):
        parsedOrder = {}
        if openOrders:
            parsedOrder["id"] = order.get('id', None)
            parsedOrder["symbol"] = order.get('market', None)
            parsedOrder["price"] = float(order["price"])
            parsedOrder["amount"] = float(order["amount"])
            side = order.get('side', None)
            if not side:
                parsedOrder["side"] = None
            else:
                parsedOrder["side"] = "buy" if side == 2 else "sell" 
            parsedOrder["timestamp"] = int(order["ctime"])*1000
            parsedOrder["status"] = "open"
            parsedOrder["orderJson"] = json.dumps(order)
        else:
            if "result" in order and order["result"]:
                parsedOrder["id"] = order["result"].get('id', None)
                parsedOrder["symbol"] = order["result"].get('market', None)
                parsedOrder["price"] = float(order["result"]["price"])
                parsedOrder["amount"] = float(order["result"]["amount"])
                side = order["result"].get('side', None)
                if not side:
                    parsedOrder["side"] = None
                else:
                    parsedOrder["side"] = "buy" if side == 2 else "sell" 
                parsedOrder["timestamp"] = int(order["result"]["ctime"])*1000
                parsedOrder["status"] = "open"
                parsedOrder["orderJson"] = json.dumps(order["result"])
        return parsedOrder

    def _parseOpenOrders(self, orders):
        parsedOrderList = []
        if "result" in orders and "records" in orders["result"]:
            for order in orders["result"]["records"]:
                parsedOrderList.append(self._parseCreateorder(order, openOrders=True))
        return parsedOrderList

    def _parseTickerInfo(self, data, ticker):
        for tickerInfo in data:
            if tickerInfo["symbol"] == self._getSymbol(ticker):
                return tickerInfo
        return {}

    ## Exchange functions 

    def create_market_sell_order(self, symbol, amount, params={}, price=None):
        apiUrl = "/api/v1/private/trade/market"
        try:
            ticker = self._getSymbol(symbol)
            params = {
                'market': self._getSymbol(ticker),
                'side': 1,
                'amount': str(amount)
            }
            params.update(params) 
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
        
    def create_market_buy_order(self, symbol, amount, params={}, price=None):
        apiUrl = "/api/v1/private/trade/market"
        try:
            ticker = self._getSymbol(symbol)
            params = {
                'market': self._getSymbol(ticker),
                'side': 2,
                'amount': amount
            }
            params.update(params) 
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e

    def create_limit_order(self, symbol, side, price, amount, params={}):
        ticker = self._getSymbol(symbol)
        if symbol not in self.symbolLotSize:
            resp = self.fetch_ticker(ticker)
            self.symbolLotSize[symbol] = resp
        # handle precisions 
        price = round(price, self.symbolLotSize[symbol]['quoteAssetPrecision'])
        amount = round(amount, self.symbolLotSize[symbol]['baseAssetPrecision'])
        
        apiUrl = "/api/v1/private/trade/limit"
        try:
            params = {
                'market': ticker,
                'side': 1 if side == "sell" else 2,
                'amount': str(amount),
                'price': str(price)
            }
            params.update(params) 
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        apiUrl = "/api/v1/private/order/pending/detail"
        try:
            ticker = self._getSymbol(symbol)
            params = {
                'market': ticker,
                'order_id': id
            }
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e

    '''
    {"code":0,"message":"Successful operation","result":{"limit":100,"offset":0,
    "records":[{"amount":"0.001","ctime":1703070656.073618,"deal_fee":"0","deal_money":"0",
    "deal_stock":"0","id":45529401586,"left":"0.001","maker_fee":"0","market":"BTC_USDT",
    "mtime":1703070656.073618,"price":"41900","side":2,"source":"api,127","taker_fee":"0",
    "type":1,"user":51340387},{"amount":"0.001","ctime":1703012664.528346,"deal_fee":"0",
    "deal_money":"0","deal_stock":"0","id":45466363178,"left":"0.001","maker_fee":"0",
    "market":"BTC_USDT","mtime":1703012664.528346,"price":"41900","side":2,"source":"api,127",
    "taker_fee":"0","type":1,"user":51340387},{"amount":"0.001","ctime":1703011778.993588,
    "deal_fee":"0","deal_money":"0","deal_stock":"0","id":45465387414,"left":"0.001","maker_fee":"0"
    ,"market":"BTC_USDT","mtime":1703011778.993588,"price":"41900","side":2,"source":"api,127",
    "taker_fee":"0","type":1,"user":51340387}],"total":3}}
    '''
    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/api/v1/private/order/pending"
        try:
            ticker = self._getSymbol(symbol)
            params = {
                # 'api_key': self.key,
                'market': self._getSymbol(ticker),
                'offset': 0,
                'limit': 100
            }
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseOpenOrders(response)
        except Exception as e:
            raise e
    
    '''
    {"code":0,"message":"Successful operation","result":{"amount":"0.001","ctime":1703071266.005806,
    "deal_fee":"0","deal_money":"0","deal_stock":"0","id":45530059627,"left":"0.001","maker_fee":"0",
    "market":"BTC_USDT","mtime":1703071266.005806,"price":"41900","side":2,"source":"api,127",
    "taker_fee":"0","type":1,"user":51340387}}
    '''
    def cancel_order(self, id, symbol=None, params={}):
        apiUrl = "/api/v1/private/trade/cancel"
        ticker = self._getSymbol(symbol)
        try:
            params = {
                'market': self._getSymbol(ticker),
                'order_id': id
            }
            response = self._send_request(method='POST', path=apiUrl, params=params)
            return self._parseCreateorder(response)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        apiUrl = "/api/v1/exchangeInfo"
        try:
            resp = self._unsignedRequest(method='GET', path=apiUrl)
            return self._parseTickerInfo(resp, symbol) # parse this response into Ticker
        except Exception as e:
            raise e
    
    # params = {"startTime": 1702512246000, "endTime": 1702512248000, "limit": 100}
    def fetch_ohlcv(self, symbol, interval, size, params={}):
       raise NotImplementedError
        
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/api/v1/private/user'
        try:
            resp = self._send_request(method='POST', path=apiUrl, params={}, payload=None)
            return self._parseBalance(resp)
        except Exception as e:
            raise e


import hmac
import base64
import hashlib
import json
import time
import requests
from ttxt.base import baseFuturesExchange
from ttxt.types import baseTypes


class gateFutures(baseFuturesExchange.BaseFuturesExchange):
    def __init__(self, key, secret, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.gateio.ws"  # host + prefix
        self.prefix = "/api/v4"
        self.settle_currency = "usdt"
        self.quanto_multiplier = {}

    def _getSymbol(self, symbol):
        return symbol.replace("/", "_")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("_", "/")

    def setQuantMultiplier(self, ticker):
        self.quanto_multiplier[ticker] = float(self.fetch_ticker(self._getSymbol(ticker))["quanto_multiplier"])

    ## Auth 
    def gen_sign(self, method, url, query_string=None, payload_string=None):
        key = self.key
        secret = self.secret
        t = time.time()
        m = hashlib.sha512()
        m.update((payload_string or "").encode('utf-8'))
        hashed_payload = m.hexdigest()
        s = '%s\n%s\n%s\n%s\n%s' % (method, url, query_string or "", hashed_payload, t)
        sign = hmac.new(secret.encode('utf-8'), s.encode('utf-8'), hashlib.sha512).hexdigest()
        return {'KEY': key, 'Timestamp': str(t), 'SIGN': sign}

    def connectToPrivateWss(self, creds, ticker):
        request = {
            "time": int(time.time()),
            "channel": "spot.orders",
            "event": "subscribe",  # "unsubscribe" for unsubscription
            "payload": [ticker]
        }
        # refer to Authentication section for gen_sign implementation
        request['auth'] = self.gen_sign(request['channel'], request['event'], request['time'], creds)
        return json.dumps(request) 

    ## Requests 
    def _signedRequest(self, request_path, method, queryString, body):
        common_headers = {'Accept': 'application/json', 'Content-Type': 'application/json'}
        if method == "GET":
            try:
                if queryString != "" and len(queryString):
                    url = self.domain_url+ self.prefix + request_path + "?" + queryString
                else:
                    url = self.domain_url+ self.prefix + request_path
                sign_headers = self.gen_sign('GET', self.prefix + request_path, queryString, "")
                sign_headers.update(common_headers)
                response = requests.get(url, headers=sign_headers)
                return response.json()
            except Exception as e:
                raise e
        if method == "POST":
            try:
                if queryString and queryString != "" and len(queryString):
                    url = self.domain_url+ self.prefix + request_path + "?" + queryString
                    sign_headers = self.gen_sign('POST', self.prefix + request_path, queryString, "")
                    sign_headers.update(common_headers)
                    response = requests.post(url, headers=sign_headers)
                else:
                    url = self.domain_url+ self.prefix + request_path
                    request_content = json.dumps(body)
                    sign_headers = self.gen_sign('POST', self.prefix + request_path, "", request_content)
                    sign_headers.update(common_headers)
                    url = self.domain_url + self.prefix + request_path
                    response = requests.post(url, headers=sign_headers, data=request_content)
                return response.json()
            except Exception as e:
                raise e
        if method == "DELETE":
            try:
                url = self.domain_url+ self.prefix + request_path
                sign_headers = self.gen_sign('DELETE', self.prefix + request_path, queryString, "")
                sign_headers.update(common_headers)
                response = requests.delete(url, headers=sign_headers)
                return response.json()
            except Exception as e:
                raise e
    
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + self.prefix + apiUrl
        if method == 'GET':
            try:
                response = requests.request('get', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")

    ## parsers
    def _parseBalance(self, balData):
        if 'message' in balData:
            raise Exception(balData['message'])
        parsedBal = {"free": {}, "total": {}, "unrealisedPnl": {}}
        parsedBal["free"][balData["currency"]] = balData.get("available", None)
        parsedBal["total"][balData["currency"]] = balData.get("total", None)
        parsedBal["unrealisedPnl"][balData["currency"]] = balData.get("unrealised_pnl", None)
        return parsedBal
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("current", None)
        parsedData["nonce"] = orderbookData.get("id", None)
        if "bids" in orderbookData:
            parsedData["bids"] = [[float(d['p']), float(d['s'])] for d in orderbookData["bids"]]
        else: parsedData["bids"] = []
        if "asks" in orderbookData:
            parsedData["asks"] = [[float(d['p']), float(d['s'])] for d in orderbookData["asks"]]
        else: parsedData["asks"] = []
        return parsedData
    
    '''
    {'stp_id': 0, 'id': 375076070136, 'biz_info': '-', 'mkfr': '0.00015', 'tkfr': '0.0004', 'tif': 'gtc', 
    'refu': 0, 'create_time': 1701883378.318, 'price': '43920.1', 'size': 10, 'status': 'open', 
    'iceberg': 0, 'is_liq': False, 'amend_text': '-', 'left': 10, 'text': 'api', 'fill_price': '0', 
    'user': 13632672, 'update_id': 1, 'refr': '0', 'is_reduce_only': False, 'stp_act': '-', 'is_close': False, 
    'contract': 'BTC_USDT'}
    '''
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('id', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["contract"]) if "contract" in order else None
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["size"]) if "size" in order else None
        if parsedOrder["amount"]:
            parsedOrder["side"] = "buy" if parsedOrder["amount"] > 0 else "sell"
        else:
            parsedOrder["side"] = None
        parsedOrder["timestamp"] = float(order["create_time"]) if "create_time" in order else None
        parsedOrder["status"] = order["status"] if "status" in order else None
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseCreateOrder(self, order):
        if 'message' in order:
            raise Exception(order['message'])
        parsedOrder = {}
        parsedOrder["id"] = order.get('id', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["contract"]) if "contract" in order else None
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["size"]) if "size" in order else None
        if parsedOrder["amount"]:
            parsedOrder["side"] = "buy" if parsedOrder["amount"] > 0 else "sell"
        else:
            parsedOrder["side"] = None
        parsedOrder["timestamp"] = float(order["create_time"]) if "create_time" in order else None
        parsedOrder["status"] = order["status"] if "status" in order else None
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseOpenOrders(self, orders):
        if 'message' in orders:
            raise Exception(orders['message'])
        parsedOrderList = []
        for order in orders:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseCandlesticks(self, candletickData):
        parsedCandlesticks = []
        for kline in candletickData:
            parsedCandlesticks.append([int(kline["t"])*1000, float(kline["o"]), float(kline["h"]), float(kline["l"]), float(kline["c"]), float(kline["v"])])
        return parsedCandlesticks

    ## Exchange functions 

    # https://www.gate.io/docs/developers/apiv4/en/#create-a-futures-order
    def create_order(self, symbol, type, side, amount, price=None, params={}):
        try:
            ticker = self._getSymbol(symbol)
            if ticker not in self.quanto_multiplier:
                self.setQuantMultiplier(ticker=ticker)
            amount = float(amount) / self.quanto_multiplier[ticker]
            amount = -1*amount if (side == 'sell') else amount # negative amount for side==sell
            if type == "market":
                body = {
                    "contract": self._getSymbol(symbol),
                    "size": amount,
                    "iceberg": params["iceberg"] if "iceberg" in params else 0,
                    "close": params["close"] if "close" in params else False,
                    "reduce_only": params["reduce_only"] if "reduce_only" in params else False,
                    "tif": params["tif"] if "tif" in params else "gtc",
                    "text": params["text"] if "text" in params else "",
                    "auto_size": params["auto_size"] if "auto_size" in params else "", #Set side to close dual-mode position. close_long closes the long side; while close_short the short one. Note size also needs to be set to 0
                    "stp_act": params["stp_act"] if "stp_act" in params else "-"
                }
            elif type == "limit":
                body = {
                    "contract": self._getSymbol(symbol),
                    "size": amount,
                    "price": float(price),
                    "iceberg": params["iceberg"] if "iceberg" in params else 0,
                    "close": params["close"] if "close" in params else False,
                    "reduce_only": params["reduce_only"] if "reduce_only" in params else False,
                    "tif": params["tif"] if "tif" in params else "gtc",
                    "text": params["text"] if "text" in params else "",
                    "auto_size": params["auto_size"] if "auto_size" in params else "", #Set side to close dual-mode position. close_long closes the long side; while close_short the short one. Note size also needs to be set to 0
                    "stp_act": params["stp_act"] if "stp_act" in params else "-"
                }
            apiUrl = f"/futures/{self.settle_currency}/orders"
            response = self._signedRequest(method='POST', request_path=apiUrl, queryString="", body=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        apiUrl = f"/futures/{self.settle_currency}/orders/{id}"
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, queryString='', body=None)
            return self._parseOrder(resp)
        except Exception as e:
            raise e

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = f"/futures/{self.settle_currency}/orders"
        try:
            query_param = 'status=open'
            resp = self._signedRequest(method='GET', request_path=apiUrl, queryString=query_param, body=None)
            return self._parseOpenOrders(resp)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None):
        apiUrl = f"/futures/{self.settle_currency}/orders/{id}"
        try:
            resp = self._signedRequest(method='DELETE', request_path=apiUrl, queryString='', body=None)
            return self._parseCreateOrder(resp)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        apiUrl = f"/futures/{self.settle_currency}/contracts/{self._getSymbol(symbol)}"
        try:
            resp = self._unsignedRequest('GET', apiUrl, params='')
            return resp # parse this response into Ticker
        except Exception as e:
            raise e
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/futures/usdt/accounts'
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, body=None, queryString=params)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
        
    def set_leverage(self, leverage, symbol, params={}):
        apiUrl = f'/futures/usdt/positions/{self._getSymbol(symbol)}/leverage'
        try:
            query_param = f'leverage={leverage}'
            resp = self._signedRequest(method='POST', request_path=apiUrl, body=None, queryString=query_param)
            return resp
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        apiUrl = f"/futures/{self.settle_currency}/order_book"
        queryParams = {"contract": self._getSymbol(symbol)}
        if limit is not None:
            queryParams["limit"] = int(limit)
        queryParams.update(params)
        try:
            resp = self._unsignedRequest('GET', apiUrl, params=queryParams)
            return self._parseOrderbook(resp)
        except Exception as e:
            raise e
    
    # parsed OHLCV = [[ts_ms, o, h ,l ,c, v],[]]
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        apiUrl = f"/futures/{self.settle_currency}/candlesticks"
        queryParams = {"contract": self._getSymbol(symbol), "from": since, "limit": limit, "interval": timeframe}
        if limit is not None:
            queryParams["limit"] = int(limit)
        queryParams.update(params)
        try:
            resp = self._unsignedRequest('GET', apiUrl, params=queryParams)
            return self._parseCandlesticks(resp)
        except Exception as e:
            raise e


from email import header
import hmac
import base64
import hashlib
import uuid
import json
from re import A
import time
from wsgiref import headers
import requests
from urllib.parse import urlencode
from ttxt.utils import general
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes

class kucoin(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, password, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.domain_url = "https://api.kucoin.com"
        self.passphrase = password
        
    def _getSymbol(self, symbol):
        return symbol.replace("/", "-")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("-", "/")
    
    def parse_params_to_str(self, params):
        paramsStr = "&".join(["%s=%s" % (key, val) for key, val in params.items()])
        url = "?" + paramsStr
        if url == '?':
            return ''
        return url
    
    # Auth
    def sign(self, message):
        mac = hmac.new(self.secret.encode('utf-8'), message.encode('utf-8'), hashlib.sha256)
        d = base64.b64encode(mac.digest())
        return d
    
    def pre_hash(self, timestamp, method, request_path, queryString, body=None):
        if not body:
            return str(timestamp) + str.upper(method) + request_path + self.parse_params_to_str(queryString)
        return str(timestamp) + str.upper(method) + request_path + body
    
    def generate_timestamp(self):
        return int(time.time() * 10**3)
    
    def _signedRequest(self, method, request_path, queryString=None, body=None):
        timeMS = self.generate_timestamp()
        if method == 'POST':
            data_json = json.dumps(body)
            queryString = json.dumps(queryString) if queryString is not None else ''
            signature = self.sign(self.pre_hash(timeMS, method, request_path, queryString='', body=data_json))
            passphrase = self.sign(self.passphrase)
            headers = {
                "KC-API-SIGN": signature,
                "KC-API-TIMESTAMP": str(timeMS),
                "KC-API-KEY": self.key,
                "KC-API-PASSPHRASE": passphrase,
                "KC-API-KEY-VERSION": "2",
                "Content-Type": "application/json" # specifying content type or using json=data in request
            }
            url = self.domain_url+request_path
            try:
                response = requests.request('POST', url, headers=headers, data=data_json)
                return response.json()
            except Exception as e:
                raise e
        if method == 'GET' or method == "DELETE":
            if queryString is None:
                queryString = ''
            signature = self.sign(self.pre_hash(timeMS, method, request_path, queryString, body=None))
            passphrase = self.sign(self.passphrase)
            headers = {
                "KC-API-SIGN": signature,
                "KC-API-TIMESTAMP": str(timeMS),
                "KC-API-KEY": self.key,
                "KC-API-PASSPHRASE": passphrase,
                "KC-API-KEY-VERSION": "2",
            }
            url = self.domain_url+request_path+self.parse_params_to_str(queryString)
            try:
                response = requests.request(method, url, headers=headers)
                return response.json()
            except Exception as e:
                raise e
        
    def _unsignedRequest(self, method, apiUrl, params):
        url = self.domain_url + apiUrl
        if method == 'GET':
            try:
                response = requests.request('GET', url, params=params)
                return response.json()
            except Exception as e:
                raise e
        else:
            raise Exception(f"{method} Method not supported for unsigned calls")
    
    ## parsers
    def _parseBalance(self, balData):
        if balData['code'] != '200000': 
            raise Exception(balData['msg'])
        parsedBal = {"free": {}, "total": {}}
        data = balData.get("data", None)
        if data is not None:
            for element in data:
                parsedBal["free"][element["currency"]] = element.get("available", None)
                parsedBal["total"][element["currency"]] = element.get("balance", None)
        return parsedBal
    
    def _parseCreateorder(self, orderData):
        if orderData['code'] != '200000': 
            raise Exception(orderData['msg'])
        parsedOrder = {'id': None, "symbol": None, "amount": None, "side": None, "timestamp": None, "status": None, "orderJson": None}
        if "data" in orderData and orderData["data"] != {}:
            parsedOrder['id'] = orderData['data']['orderId']
            parsedOrder["orderJson"] = json.dumps(orderData)
        return parsedOrder
    
    def _parseCancelorder(self, orderData):
        if orderData['code'] != '200000': 
            raise Exception(orderData['msg'])
        parsedOrder = {'id': None, "symbol": None, "amount": None, "side": None, "timestamp": None, "status": None, "orderJson": None}
        if "data" in orderData and "cancelledOrderIds" in orderData["data"] and type(orderData['data']['cancelledOrderIds']) == list:
            if not orderData['data']['cancelledOrderIds']:
                return parsedOrder
            else:
                parsedOrder['id'] = orderData['data']['cancelledOrderIds'][0]
        return parsedOrder
    
    def _parseOpenOrders(self, orderjson):
        parsedOrderList = []
        if "data" in orderjson and "items" in orderjson["data"]:
            for order in orderjson["data"]["items"]:
                parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('id', None)
        parsedOrder["symbol"] = self._getUserSymbol(order.get('symbol', None))
        parsedOrder["price"] = float(order["price"]) if "price" in order else None 
        parsedOrder["amount"] = float(order["size"]) if "size" in order else None
        parsedOrder["side"] = order['side'].lower() if "side" in order else None
        parsedOrder["timestamp"] = order["createdAt"] if "cTime" in order else None
        parsedOrder["status"] = order.get("status", None)
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseFetchTrades(self, orderData):
        if orderData['code'] != '200000': 
            raise Exception(orderData['msg'])
        parsedTradesList = []
        if "data" in orderData and "items" in orderData["data"]:
            for order in orderData["data"]['items']:
                parsedTradesList.append(self._parseTrades({'data': order}))
        return parsedTradesList
        
    def _parseTrades(self, data):
        parsedOrder = {"id": None, "symbol": None, "price": None, "amount": None, "takerOrMaker": None, "datetime": None, "fee": None, "fee_currency": None,
                       "side": None, "timestamp": None, "status": None}
        if "data" in data and data["data"] != {}:
            order = data['data']
            if type(order) == list:
                order = order[0]
            parsedOrder["id"] = order.get("orderId", None)
            parsedOrder["tradeId"] = order.get("tradeId", None)
            parsedOrder["symbol"] = self._getUserSymbol(order.get("symbol", None))
            if "price" in order:
                parsedOrder["price"] = float(order.get("price", None))
            else:
                parsedOrder["price"] = float(order.get("priceAvg", None))
            parsedOrder['takerOrMaker'] = order.get('liquidity', None)
            parsedOrder["amount"] = float(order.get("size", None))
            parsedOrder["side"] = order.get("side").lower()
            parsedOrder["timestamp"] = int(order["createdAt"])
            parsedOrder["datetime"] = general.ts_to_datetime(parsedOrder["timestamp"])
            parsedOrder["status"] = order.get('status', None)
            parsedOrder["fee_currency"] = order.get('feeCurrency', None)
            parsedOrder['fee'] = order.get("fee", None)
        return parsedOrder
    
    ## Exchange functions 
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = "/api/v1/accounts"
        try:
            params = { "type": "trade" }
            response = self._signedRequest('GET', request_path=apiUrl, queryString=params)
            return self._parseBalance(response)
        except Exception as e:
            raise e
    
    def fetch_my_trades(self, symbol, since=None, limit=None, params={}): 
        apiUrl = "/api/v1/fills"
        try:
            params = {
                "symbol": self._getSymbol(symbol),
                "startAt": '0',
                "endAt": str(self.generate_timestamp())
            }
            if since is not None:
                params['startAt'] = since
            if 'endTime' in params and params['endTime']:
                params['endAt'] = params['endTime']
            response = self._signedRequest('GET', request_path=apiUrl, queryString=params, body='')
            return self._parseFetchTrades(response)
        except Exception as e:
            raise e
    
    def create_order(self, symbol, side, amount, order_type, price=None, params={}): # change body
        apiUrl = "/api/v1/orders"
        try:
            params = {
                "clientOid": str(uuid.uuid4()),
                "side": side,
                "symbol": self._getSymbol(symbol),
                "size": amount
            }
            if order_type == 'limit':
                params['type'] = 'limit'
                params['timeInForce'] = 'GTC'
                params["price"] = price
            if order_type == 'market':
                params['type'] = 'market'
            response = self._signedRequest('POST', request_path=apiUrl, body=params, queryString='')
            return self._parseCreateorder(response)
        except Exception as e:
            raise e
    
    def create_limit_order(self, symbol, side, amount, price, order_type=None, params={}):
        return self.create_order(symbol, side, amount=amount, price=price, order_type='limit', params={})
    
    def create_market_order(self, symbol, side, amount, order_type=None, price=None, params={}):
        return self.create_order(symbol, side, amount, order_type='market', price=None, params={})
    
    def cancel_order(self, id, symbol=None, params={}):
        apiUrl = "/api/v1/orders"
        try:
            params = {
                'orderId': id
            }
            response = self._signedRequest("DELETE", request_path=apiUrl, queryString=params)
            return self._parseCancelorder(response)
        except Exception as e:
            raise e
    
    def fetch_open_orders(self, symbol=None, kwargs=None):
        apiUrl = '/api/v1/orders'
        ticker = self._getSymbol(symbol)
        try:
            params = {
                'status': "active"
            }
            if symbol is not None:
                params['symbol'] = ticker
            response = self._signedRequest("GET", request_path=apiUrl, queryString=params, body='')
            return self._parseOpenOrders(response)
        except Exception as e:
            raise e


from ttxt.base import baseFuturesExchange, baseSpotExchange
from ttxt.exchanges.gateFutures import gateFutures
from ttxt.exchanges.bybitFutures import bybitFutures
from ttxt.exchanges.bitgetFutures import bitgetFutures
from ttxt.exchanges.bingx import bingx
from ttxt.exchanges.biconomy import biconomy
from ttxt.exchanges.mexc import mexc
from ttxt.exchanges.gateio import gateio
from ttxt.exchanges.bybit import bybit
from ttxt.exchanges.okx import okx
from ttxt.exchanges.huobi import huobi
from ttxt.exchanges.bitmart import bitmart
from ttxt.exchanges.cryptocom import cryptocom
from ttxt.exchanges.bitget import bitget
from ttxt.exchanges.binance import binance
from ttxt.exchanges.kucoin import kucoin

exchanges = [
    "biconomy",
    "binance",
    "bingx",
    "bitgetFutures",
    "bitget",
    "bybitFutures",
    "gateFutures",
    "kucoin"
    "mexc",
    "gateio",
    "bybit",
    "okx",
    "huobi",
    "bitmart",
    "cryptocom"
]

base = [
    "baseFuturesExchange",
    "baseSpotExchange"
]

_all__ =  exchanges + base


import hmac
import base64
import hashlib
import json
import time
import requests
import datetime
from ttxt.base import baseSpotExchange
from ttxt.types import baseTypes
from ttxt.utils import general


class okx(baseSpotExchange.BaseSpotExchange):
    def __init__(self, key, secret, password, **kwargs):
        super().__init__(key, secret, **kwargs)
        self.password = password
        self.domain = "https://www.okx.com"
        ## constants
        self.CONTENT_TYPE = 'Content-Type'
        self.OK_ACCESS_KEY = 'OK-ACCESS-KEY'
        self.OK_ACCESS_SIGN = 'OK-ACCESS-SIGN'
        self.OK_ACCESS_TIMESTAMP = 'OK-ACCESS-TIMESTAMP'
        self.OK_ACCESS_PASSPHRASE = 'OK-ACCESS-PASSPHRASE'
        self.APPLICATION_JSON = 'application/json'
        self.GET = "GET"
        self.POST = "POST"

    def _getSymbol(self, symbol):
        return symbol.replace("/", "-")
    
    def _getUserSymbol(self, symbol):
        return symbol.replace("-", "/")

    ## Auth 
    def sign(self, message):
        mac = hmac.new(bytes(self.secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return base64.b64encode(d)


    def pre_hash(self, timestamp, method, request_path, body):
        return str(timestamp) + str.upper(method) + request_path + body


    def get_header(self, sign, timestamp):
        header = dict()
        header[self.CONTENT_TYPE] = self.APPLICATION_JSON
        header[self.OK_ACCESS_KEY] = self.key
        header[self.OK_ACCESS_SIGN] = sign
        header[self.OK_ACCESS_TIMESTAMP] = str(timestamp)
        header[self.OK_ACCESS_PASSPHRASE] = self.password
        # header['x-simulated-trading'] = flag
        return header

    def get_header_no_sign(self):
        header = dict()
        header[self.CONTENT_TYPE] = self.APPLICATION_JSON
        # header['x-simulated-trading'] = flag
        return header

    def parse_params_to_str(self, params):
        url = '?'
        for key, value in params.items():
            if(value != ''):
                url = url + str(key) + '=' + str(value) + '&'
        url = url[0:-1]
        return url

    def get_timestamp(self):
        now = datetime.datetime.utcnow()
        t = now.isoformat("T", "milliseconds")
        return t + "Z"

    def signature(self, timestamp, method, request_path, body):
        if str(body) == '{}' or str(body) == 'None':
            body = ''
        message = str(timestamp) + str.upper(method) + request_path + str(body)

        mac = hmac.new(bytes(self.secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()

        return base64.b64encode(d)

    ## Requests 
    def _signedRequest(self, method, request_path, params):
        if method == "GET":
            request_path = request_path + self.parse_params_to_str(params)
        timestamp = self.get_timestamp()
        body = json.dumps(params) if method == "POST" else ""
        if self.key != '-1':
            sign = self.sign(self.pre_hash(timestamp, method, request_path, str(body)))
            header = self.get_header(sign, timestamp)
        else:
            header = self.get_header_no_sign()
        response = None
        url = self.domain + request_path
        if method == "GET":
            response = requests.get(url, headers=header)
        elif method == "POST":
            response = requests.post(url, data=body, headers=header)
        return response.json()
    
    ## parsers
    def _parseBalance(self, balData):
        if balData['code'] != "0": raise Exception(balData["msg"])
        parsedBal = {"free": {}, "total": {}}
        for bal in balData["data"][0]["details"]:
            parsedBal["free"][bal["ccy"]] = bal.get("availBal", None)
            parsedBal["total"][bal["ccy"]] = bal.get("eq", None)
        return parsedBal
    
    def _parseOrderbook(self, orderbookData):
        parsedData = {}
        parsedData["datetime"] = None
        parsedData["timestamp"] = orderbookData.get("current", None)
        parsedData["nonce"] = orderbookData.get("id", None)
        if "bids" in orderbookData:
            parsedData["bids"] = [[float(d[0]), float(d[1])] for d in orderbookData["bids"]]
        else: parsedData["bids"] = []
        if "asks" in orderbookData:
            parsedData["asks"] = [[float(d[0]), float(d[1])] for d in orderbookData["asks"]]
        else: parsedData["asks"] = []
        return parsedData

    def _parseCreateOrder(self, order):
        parsedOrder = {}
        if order['code'] != "0": raise Exception(order["msg"])
        parsedOrder["id"] = order['data'][0]['ordId']
        parsedOrder["symbol"] = None
        parsedOrder["price"] = None
        parsedOrder["amount"] = None
        parsedOrder["side"] = None
        parsedOrder["timestamp"] = float(order["outTime"]) if "outTime" in order else None
        parsedOrder["status"] = None
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder
    
    def _parseOrder(self, order):
        parsedOrder = {}
        parsedOrder["id"] = order.get('ordId', None)
        parsedOrder["symbol"] = self._getUserSymbol(order["instId"]) if "instId" in order else None
        parsedOrder["price"] = float(order["px"]) if "px" in order else None 
        parsedOrder["amount"] = float(order["sz"]) if "sz" in order else None
        parsedOrder["side"] = order.get('side', None)
        parsedOrder["timestamp"] = float(order["cTime"]) if "cTime" in order else None
        parsedOrder["status"] = order["status"] if "status" in order else None
        parsedOrder["orderJson"] = json.dumps(order) if order else None 
        return parsedOrder

    def _parseOpenOrders(self, orders):
        parsedOrderList = []
        if order['code'] != "0": raise Exception(order["msg"])
        for order in orders['data']:
            parsedOrderList.append(self._parseOrder(order))
        return parsedOrderList
    
    def _parseTrades(self, trades):
        def parseTrade(trade):
            parsedTrade = {}
            parsedTrade["symbol"] = trade.get('instId', None)
            parsedTrade["id"] = trade.get("ordId", None)
            parsedTrade["tradeId"] = trade.get("tradeId", None)
            parsedTrade["side"] = trade.get("side", None)
            parsedTrade["price"] = trade.get("fillPx", None)
            parsedTrade["amount"] = trade.get("fillSz", None)
            parsedTrade["takerOrMaker"] = "taker" if trade['execType'] == 'T' else "maker"
            if "fillTime" in trade:
                parsedTrade["timestamp"] = int(float(trade["fillTime"]))
                parsedTrade["datetime"] = general.ts_to_datetime(parsedTrade["timestamp"])
            else: 
                parsedTrade["timestamp"] = None
                parsedTrade["datetime"] = None
            parsedTrade["feeCurrency"] = trade["feeCcy"]
            parsedTrade["fee"] = trade["fee"]
            return parsedTrade
        parsedTradeList = []
        if trades['code'] != "0": raise Exception(trades["msg"])
        for trade in trades["data"]:
            parsedTradeList.append(parseTrade(trade))
        return parsedTradeList

    ## Exchange functions 

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        if type == "market":
            body = {
                "tdMode": "cash",
                "instId": self._getSymbol(symbol),
                "ordType": "market",
                "side": side,
                "iceberg": params["iceberg"] if "iceberg" in params else "0",
                "sz": str(amount)
            }
        elif type == "limit":
            body = {
                "tdMode": "cash",
                "instId": self._getSymbol(symbol),
                "ordType": "limit",
                "side": side,
                "iceberg": params["iceberg"] if "iceberg" in params else "0",
                "sz": str(amount),
                "px": str(price)
            }
        apiUrl = f"/api/v5/trade/order"
        try:
            response = self._signedRequest(method='POST', request_path=apiUrl, params=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e
        
    def create_limit_order(self, symbol, amount, side, price, params={}):
        body = {
            "tdMode": "cash",
            "instId": self._getSymbol(symbol),
            "ordType": "limit",
            "side": side,
            "iceberg": params["iceberg"] if "iceberg" in params else "0",
            "sz": str(amount),
            "px": str(price)
        }
        apiUrl = f"/api/v5/trade/order"
        try:
            response = self._signedRequest(method='POST', request_path=apiUrl, params=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e

    def create_market_order(self, symbol, amount, side, params={}):
        body = {
            "tdMode": "cash",
            "instId": self._getSymbol(symbol),
            "ordType": "market",
            "side": side,
            "iceberg": params["iceberg"] if "iceberg" in params else "0",
            "sz": str(amount)
        }
        apiUrl = f"/api/v5/trade/order"
        try:
            response = self._signedRequest(method='POST', request_path=apiUrl, params=body)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e
    
    def fetch_order(self, id, symbol=None):
        raise NotImplementedError("method not implemented")

    def fetch_open_orders(self, symbol=None,  kwargs=None):
        apiUrl = "/api/v5/trade/orders-pending"
        params = {"instId": self._getSymbol(symbol), "instType": "SPOT"}
        try:
            response = self._signedRequest(method='GET', request_path=apiUrl, params=params)
            return self._parseOpenOrders(response)
        except Exception as e:
            raise e
        
    def cancel_order(self, id, symbol=None):
        apiUrl = "/api/v5/trade/cancel-order"
        params = {"instId": self._getSymbol(symbol), "ordId": id}
        try:
            response = self._signedRequest(method='POST', request_path=apiUrl, params=params)
            return self._parseCreateOrder(response)
        except Exception as e:
            raise e

    def fetch_ticker(self, symbol: str, params={}) -> baseTypes.Ticker:
        raise NotImplementedError("method not implemented")
    
    def fetch_balance(self, params={}) -> baseTypes.Balances:
        apiUrl = '/api/v5/account/balance'
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, params=params)
            return self._parseBalance(resp)
        except Exception as e:
            raise e
    
    # for extra params: https://www.okx.com/docs-v5/en/#order-book-trading-trade-get-transaction-details-last-3-months
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        apiUrl = '/api/v5/trade/fills-history'
        if symbol:
            params['instType'] = "SPOT"
        if limit:
            params['limit'] = limit
        if since:
            params['begin'] = int(since)
        try:
            resp = self._signedRequest(method='GET', request_path=apiUrl, params=params)
            return self._parseTrades(resp)
        except Exception as e:
            raise e
        
    def fetch_order_book(self, symbol: str, limit=None, params={}) -> baseTypes.OrderBook:
        raise NotImplementedError("method not implemented")
    
    # parsed OHLCV = [[ts_ms, o, h ,l ,c, v],[]]
    def fetch_ohlcv(self, symbol: str, timeframe='1m', since = None, limit = None, params={}):
        raise NotImplementedError("method not implemented")